#ifndef GRAD_H
#define GRAD_H 1

#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "Clean.h"
#include <omp.h>
typedef long long llong;
#define C1 0.125
#define C2 0.25
/*
	Created by Yinbin Ma  
	4th order Laplacian in 2 d.
	Use carefully and have fun.
	
	First created: 2015-01-01
	Last Modified: 2015-02-28
*/	

namespace SEP {

class GradX: public Operator{
  public:
    GradX(Space *mod, Space *data, llong _nb){
      setDomain(mod);
      setRange(data);
      nb = _nb;
      init();
    }
    ~GradX(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    private:
    void init();
    float *g;
    double pi;
    llong n1, n2;
    llong i1, i2, position;
    llong nb;
    float temp0,temp1,temp2;
    CleanVec Zero;

};

void GradX::init() {
  pi = 3.1415926535;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n1 = in_dims[0];
  n2 = in_dims[1];
  g = new float[nb*2];
  float flag = pow(-1.,nb+1);
  for(llong i=-nb; i < nb; i++){
    g[i+nb] = flag/pi/(i+0.5)/(i+0.5);
    flag = -flag;
    std::cout << " at i = " << i<<", g = " << g[i+nb] <<std::endl;
  }
}

bool GradX::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, lag, position;
 //   #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-2; i2++){
      position = i1*n2+i2;
      for(lag = -nb; lag < nb; lag++){
        d[position] += g[lag+nb]*m[position-lag];
      }

    }
  }



  return 0;
}

bool GradX::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, lag, position;
 //   #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-2; i2++){
      position = i1*n2+i2;
      for(lag = -nb; lag < nb; lag++){
         m[position-lag]+= g[lag+nb]*d[position];
      }
    }
  }

  return 0;
}
class GradZ: public Operator{
  public:
    GradZ(Space *mod, Space *data, llong _nb){
      setDomain(mod);
      setRange(data);
      nb = _nb;
      init();
    }
    ~GradZ(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    private:
    void init();
    float *g;
    float a0, a1, a2;
    llong n1, n2;
    llong i1, i2, position;
    llong nb;
    float temp0,temp1,temp2;
    CleanVec Zero;

};

void GradZ::init() {
  pi = 3.1415926535;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n1 = in_dims[0];
  n2 = in_dims[1];
  g = new float[nb*2];
  float flag = pow(-1.,nb+1);
  for(llong i=-nb; i < nb; i++){
    g[i+nb] = flag/pi/(i+0.5)/(i+0.5);
    flag = -flag;
  }
}

bool GradZ::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
    // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, lag, position;
   // #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-2; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      position = i1*n2+i2;
      for(lag = -nb; lag < nb; lag++){
        d[position] += g[lag+nb]*m[position-lag*nx];
      }
    }
  }

  return 0;
}

bool GradZ::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, lag, position;
  for(i1 = nb; i1 < n1-nb-2; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      for(lag = -nb; lag < nb; lag++){
        d[position] += g[lag+nb]*m[position-lag*nx];
      }
    
    }
  }

  return 0;
}

class GradX2: public Operator{
  public:
    GradX2(Space *mod, Space *data, llong _nb){
      setDomain(mod);
      setRange(data);
      nb = _nb;
      init();
    }
    ~GradX2(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    private:
    void init();
    float *g;
    float a0, a1, a2;
    llong n1, n2;
    llong i1, i2, position;
    llong nb;
    float temp0,temp1,temp2;
    CleanVec Zero;

};

void GradX2::init() {
  pi = 3.1415926535;
  a0 = 0.5;
  a1 = 9./8.;
  a2 = -1./24.;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n1 = in_dims[0];
  n2 = in_dims[1];
  g = new float[nb*2];
  float flag = pow(-1.,nb+1);
  for(llong i=-nb; i < nb; i++){
    g[i+nb] = flag/pi/(i+0.5)/(i+0.5);
    flag = -flag;
  }
}

bool GradX2::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, lag, position;
  //  #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      position = i1*n2+i2;
      for(lag = -nb; lag < nb; lag++){
        d[position] += g[lag+nb]*m[position-(lag+1)];
      }
    }
  }



  return 0;
}

bool GradX2::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, lag, position;
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      position = i1*n2+i2;
      for(lag = -nb; lag < nb; lag++){
        m[position-(lag+1)] += g[lag+nb]*d[position];
      }
    }
  }

  return 0;
}
class GradZ2: public Operator{
  public:
    GradZ2(Space *mod, Space *data, llong _nb){
      setDomain(mod);
      setRange(data);
      nb = _nb;
      init();
    }
    ~GradZ2(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    private:
    void init();
    float *g;
    float a0, a1, a2;
    llong n1, n2;
    llong i1, i2, position;
    llong nb;
    float temp0,temp1,temp2;
    CleanVec Zero;

};

void GradZ2::init() {
  pi = 3.1415926535;
  a0 = 0.5;
  a1 = 9./8.;
  a2 = -1./24.;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n1 = in_dims[0];
  n2 = in_dims[1];
  g = new float[nb*2];
  float flag = pow(-1.,nb+1);
  for(llong i=-nb; i < nb; i++){
    g[i+nb] = flag/pi/(i+0.5)/(i+0.5);
    flag = -flag;
  }
}

bool GradZ2::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
    // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, lag, position;
  //  #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      position = i1*n2+i2;
      for(lag = -nb; lag < nb; lag++){
        d[position] += g[lag+nb]*m[position-(lag+1)*nx];
      }
    }
  }

  return 0;
}

bool GradZ2::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, lag, position;
   // #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      position = i1*n2+i2;
      for(lag = -nb; lag < nb; lag++){
        m[position-(lag+1)*nx] += g[lag+nb]*d[position];
      }
    }
  }

  return 0;
}


}
#endif