#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"

#include <omp.h>
#include "base/utilities.h"

#include "base/s_reg.h"
#include "base/hypercube.h"
#include "base/Vector.h"
#include "base/inCoreFloat.h"
#include "base/DataInCoreFloat.h"
#include "base/Operator.h"
#include "base/MultiOperator.h"

#include "Interpolation.h"
#include "CreateModel.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Prop.h"
#include "Imaging.h"

extern "C" {
#include "sep3d.h"
}
using namespace std;
using namespace SEP;

/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/
extern "C"  int sep_get_data_axis_par(const char *r,int *ax,int *ff,float *o,float *d,char *l);

int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "         Full Waveform inversion.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(4);   // Initialize the parallel environment
  
  // Define dimension of interest
  int nx, nz, nt;
  int sx, sz,ns,oxs,dxs;
  int boundary;
  float dx, dz, dt;
  float velocity, density, modulus;
  float fmax=15;

  // Get parameters
  getch("nx","d",&nx);    cout << "	nx = " << nx << endl;
  getch("nz","d",&nz);    cout << "	nz = " << nz << endl;
  getch("nt","d",&nt);    cout << "	nt = " << nt << endl;
  getch("boundary","d",&boundary);    cout << "	boundary = " << boundary << endl;
  getch("dx","f",&dx);    cout << "	dx = " << dx <<"[m]"<< endl;
  dz = dx;
  getch("sz","d",&sz);    cout << "	sz = " << sz << endl;
  getch("oxs","d",&oxs);  cout << "	oxs = " << oxs << endl;
  getch("dxs","d",&dxs);  cout << "	dxs = " << dxs << endl;
  getch("dt","f",&dt);    cout << "	dt = " << dt << "[ms]"<< endl;
  getch("ns","d",&ns);    cout << "	ns = " << ns << endl;
  getch("density","f",&density); cout << "	density = " << density << "[g/cm^3]" <<endl;
  getch("velocity","f",&velocity); cout << "	velocity = " << velocity << "[km/s]" <<endl;
  modulus=velocity*velocity*density;  cout << "	modulus = " << modulus << "[GPa]" <<endl; 

  // Set up space 
  inCoreFloat1D sspace(nt);               // space for shot
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wave field cube
  inCoreFloat2D dspace(nt,nx);            // space for data space

  // Set up vectors
  Vector *K      = new Vector(&mspace);    K->zeroElement();      // migration velocity
  Vector *rho    = new Vector(&mspace);    rho->zeroElement();     // Velocity vector
  Vector *K0     = new Vector(&mspace);    K0->zeroElement();      // migration velocity
  Vector *rho0   = new Vector(&mspace);    rho0->zeroElement();     // Velocity vector
  Vector *dmK    = new Vector(&mspace);    dmK->zeroElement();      // migration velocity
  Vector *dmrho  = new Vector(&mspace);    dmrho->zeroElement();     // Velocity vector
  Vector *GradK  = new Vector(&mspace);    GradK->zeroElement();      // migration velocity
  Vector *Gradr  = new Vector(&mspace);    Gradr->zeroElement();     // Velocity vector
  Vector *vel    = new Vector(&mspace);    vel->zeroElement();      // migration velocity
  Vector *shot   = new Vector(&sspace);    shot->zeroElement();     // shot profile
  Vector *data   = new Vector(&dspace);    data->zeroElement();     // data
  Vector *us     = new Vector(&uspace);    us->zeroElement();       // Wave field
  Vector *u0     = new Vector(&uspace);    u0->zeroElement();       // Wave field
  Vector *u1     = new Vector(&uspace);    u1->zeroElement();       // Wave field
  Vector *borns  = new Vector(&uspace);    borns->zeroElement();       // Wave field
  Vector *ub     = new Vector(&uspace);    ub->zeroElement();       // Wave field
  
  
  // Read in wavelet and construct model
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  sreed("wavelet",shot_val,sizeof(float)*nt); 
    
  // Set up model
  Models *CModel =  new Models(nz, nx, modulus, density);
  CModel->ModelC(K0,rho0);
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();
  float *__restrict__ K0_val = ((dataInCoreFloat*)K0->getDataContainer())->getFloatPtr(); 
  float *__restrict__ dmK_val = ((dataInCoreFloat*)dmK->getDataContainer())->getFloatPtr(); 
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr(); 
  float *__restrict__ rho0_val = ((dataInCoreFloat*)rho0->getDataContainer())->getFloatPtr();   
  float *__restrict__ dmrho_val = ((dataInCoreFloat*)dmrho->getDataContainer())->getFloatPtr();  
  K->linearCombo(0.,*K0,1.);
  rho->linearCombo(0.,*rho0,1.);
  K_val[150*nx+110] = K_val[150*nx+110]*1.01;
  dmK_val[150*nx+110] = -0.01/K0_val[150*nx+110];
  rho_val[150*nx+190] = rho_val[150*nx+190]*1.01;
  dmrho_val[150*nx+190] = -0.01/rho0_val[150*nx+190];


  // Check if delta_model is computed accurately
  IntX *interpX = new IntX(&mspace,&mspace);
  Vector *mrho = new Vector(&mspace);
  Vector *temp = new Vector(&mspace);
  Vector *tdmrho = new Vector(&mspace);
  float *__restrict__ mrho_val = ((dataInCoreFloat*)mrho->getDataContainer())->getFloatPtr();
  float *__restrict__ temp_val = ((dataInCoreFloat*)temp->getDataContainer())->getFloatPtr();
  float *__restrict__ tdmrho_val = ((dataInCoreFloat*)tdmrho->getDataContainer())->getFloatPtr(); 
  for(long long position=0; position<nx*nz;position++){
	temp_val[position] = 1./rho_val[position];
  }
  interpX->Forward(false,temp,tdmrho); tdmrho->addInv();
  for(long long position=0; position<nx*nz;position++){
	temp_val[position] = 1./rho0_val[position];
  }  
  interpX->Forward(true,temp,tdmrho); 
  double norm1=tdmrho->innerProd(*tdmrho);
  srite("dmrho",tdmrho_val,sizeof(float)*nx*nz);
  interpX->Forward(false,dmrho,tdmrho);
  double norm2=tdmrho->innerProd(*tdmrho);
  cout << "relative error = " << (norm1-norm2)/norm1 <<endl;


   
  // Check numerical condition
  cout << "    Dispersion: Vmin/(fmax*dx) = " << 1000.*velocity/fmax/dx <<", should be > 4"<<endl;
  cout << "    Courant:    Vmax*dt/dx     = " << velocity*dt/dx <<", should be < 0.717"<<endl;
 
  
  // Set up operators
  SensorPad *SP = new SensorPad(&dspace, &uspace, sz, boundary);
  Gradient *Imag = new Gradient(u1, K, boundary, dt, dx);
  Prop *Lmig = new Prop(&uspace,&uspace);    // propogation to generate data
  Prop *Lmig0= new Prop(&uspace,&uspace);    // propagation using migration velocity
  Lmig->SetPropPar(boundary,dt, dx, K);
  Lmig0->SetPropPar(boundary,dt, dx, K);
  Lmig->SetModel(K,rho);
  Lmig0->SetModel(K0,rho0);
  Source *Ss = new Source(nt, nz, nx);  
  

  // Inject source wavelet
  cout <<"sx = " << sx <<endl;
  sx = oxs;
  sx = 150; sz=90;
  Ss->Injection(shot, us, sz, sx);

  // Forward propagation
  double t0 = timer();
  Lmig0->Forward(false,us,u0);        // background pressure
  Lmig->Forward(false,us,u1);         // true pressure
  u1->linearCombo(1.,*u0,-1.);        // true difference
  Imag->BornS(dmK,dmrho, u0, borns);  // Born approximation
  std::cout << "born norm = " << borns->innerProd(*borns) << endl;
  Lmig0->Forward(false,borns,ub);     // difference from Born approximation
  double t1 = timer(); 
  cout << "   wave propagation, takes " << t1-t0 << "seconds. " <<endl;
  
  


  // output result
  char l[128]="label";
  int axis;
  float o, dtout;
  o = 0; dtout=dt/1000;
  float *__restrict__ u1_val = ((dataInCoreFloat*)u1->getDataContainer())->getFloatPtr();
  double pnormt=sqrt(u1->innerProd(*u1));
  cout << "true norm = " << pnormt <<endl;
  srite("truewave",u1_val,sizeof(float)*nt*nz*nx);
  axis=3;
  sep_put_data_axis_par("truewave",&axis,&nt,&o,&dtout,l);
  axis=2;
  sep_put_data_axis_par("truewave",&axis,&nz,&o,&dz,l);
  axis=1;
  sep_put_data_axis_par("truewave",&axis,&nx,&o,&dx,l);
  float *__restrict__ u2_val = ((dataInCoreFloat*)ub->getDataContainer())->getFloatPtr();
  double pnormb=sqrt(ub->innerProd(*ub));
  cout << "born norm = " << sqrt(ub->innerProd(*ub)) <<endl;
  srite("bornwave",u2_val,sizeof(float)*nt*nz*nx);
  axis=3;
  sep_put_data_axis_par("bornwave",&axis,&nt,&o,&dtout,l);
  axis=2;
  sep_put_data_axis_par("bornwave",&axis,&nz,&o,&dz,l);
  axis=1;
  sep_put_data_axis_par("bornwave",&axis,&nx,&o,&dx,l);
  ub->linearCombo(1.,*u1,-1.);
  double pnormd=sqrt(ub->innerProd(*ub));
  cout << "diff norm = " <<  pnormd <<endl;
  cout << "relative error = " << pnormd/pnormt <<endl;
  srite("diffwave",u2_val,sizeof(float)*nt*nz*nx);
  axis=3;
  sep_put_data_axis_par("diffwave",&axis,&nt,&o,&dtout,l);
  axis=2;
  sep_put_data_axis_par("diffwave",&axis,&nz,&o,&dz,l);
  axis=1;
  sep_put_data_axis_par("diffwave",&axis,&nx,&o,&dx,l);
  // Clean up

   std::cout << "========== Done, Cleaning up ==========" <<std::endl; 

  return 0;
}