#ifndef CREATEMODEL_H
#define CREATEMODEL_H 1
#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "base/Vector.h"
#include <cmath>
#include <omp.h>
#include <iostream>
typedef long long llong;
/*
Model creation module
Used to simplify the main code.

Created:       2015-07-06
Last modified: 2015-07-06

*/

namespace SEP {

class Models{
  
  public:
    Models(llong _nz, llong _nx, float _modulus, float _density){
      nz = _nz;
      nx = _nx;
      modulus = _modulus;
      density = _density;
    }
    ~Models(){}
    void ModelC(Vector *K, Vector *rho);
    void Model0(Vector *K, Vector *rho);
    void Model1(Vector *K, Vector *rho);
    void Model2(Vector *K, Vector *rho);
    void Model6(Vector *K, Vector *rho);
    void Model7(Vector *K, Vector *rho);
    void Model8(Vector *K, Vector *rho);
    void Model9(Vector *K, Vector *rho);
  private:
    llong nz,nx;
    float modulus, density;
};

void Models::ModelC(Vector *K, Vector *rho) {
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();   
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr();  

   
  for(llong i1=0; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus;
      rho_val[i] = density;
    }
  }    

}
void Models::Model0(Vector *K, Vector *rho) {
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();   
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr();  

  for(llong i1=0; i1<110;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus;
      rho_val[i] = density*1.0;
    }
  }   

  for(llong i1=110; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus;
      rho_val[i] = density*1.2;
    }
  }     
}

void Models::Model1(Vector *K, Vector *rho) {
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();   
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr();  

  for(llong i1=0; i1<110;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus;
      rho_val[i] = density*1.0;
    }
  }   

  for(llong i1=110; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus/1.2;
      rho_val[i] = density;
    }
  }    
}

void Models::Model2(Vector *K, Vector *rho) {
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();   
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr();  

   
  for(llong i1=0; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus;
      rho_val[i] = density;
    }
  } 
  for(llong i1=135; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus*0.8;
      rho_val[i] = density*1.1;
    }
  } 
}
void Models::Model6(Vector *K, Vector *rho) {
  // Model that contains a Gaussian anamoly centered around (200,130)

  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();   
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr();  

   
  for(llong i1=0; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus;
      rho_val[i] = density;
    }
  } 
  for(llong i1=135; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus*0.8;
      rho_val[i] = density*1.1;
    }
  } 
  //rho_val[110*nx+150] = density*1.01;
  //K_val[110*nx+150] = modulus*0.99;
  
   for(llong i1=90; i1<110;i1++){
    for(llong i2=220; i2<280; i2++){
      llong i = i1*nx+i2;
        K_val[i] = K_val[i]-0.05*modulus; //*exp(-1.*distance/200.);
        rho_val[i] = rho_val[i]+0.03*density; //*exp(-1.*distance/200.);
    }
  } 
  
 /*
  for(llong i1=0; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      float distance = (i2-150.)*(i2-150.)+(i1-100.)*(i1-100.);
      if(distance < 800.){
      //  rho_val[i] = rho_val[i]+0.01*density; //*exp(-1.*distance/200.);
        K_val[i] = K_val[i]-0.1*modulus*exp(-1.*distance/100.);
      }
    }
  }
  */
}  
void Models::Model7(Vector *K, Vector *rho) {
  // Model that contains a Gaussian anamoly centered around (200,130)

  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();   
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr();  

   
  for(llong i1=0; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus;
      rho_val[i] = density;
    }
  } 
  for(llong i1=135; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus*0.8;
    //  rho_val[i] = density*1.1;
    }
  } 
  
      llong i = 100*nx+150;
        K_val[i] = K_val[i]-0.02*modulus; //*exp(-1.*distance/200.);

  
 /*
  for(llong i1=0; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      float distance = (i2-150.)*(i2-150.)+(i1-100.)*(i1-100.);
      if(distance < 800.){
      //  rho_val[i] = rho_val[i]+0.01*density; //*exp(-1.*distance/200.);
        K_val[i] = K_val[i]-0.1*modulus*exp(-1.*distance/100.);
      }
    }
  }
  */
}  

void Models::Model8(Vector *K, Vector *rho) {
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();   
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr();  

  llong z0,z1,z2,z3,z4,z5;
  z0=0; z1=80; z2=110; z3=130; z4=148;
   
  for(llong i1=0; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus;
      rho_val[i] = density;
    }
  } 
  for(llong i1=z1; i1<z2;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus*1.1;
    }
  } 
  for(llong i1=z3; i1<z4;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus*1.1;
    }
  }
}

void Models::Model9(Vector *K, Vector *rho) {
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();   
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr();  

  llong z0,z1,z2,z3,z4,z5;
  z0=0; z1=80; z2=110; z3=130; z4=148;
   
  for(llong i1=0; i1<nz;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus;
      rho_val[i] = density;
    }
  } 
  for(llong i1=z1; i1<z2;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus*1.1;
    }
  } 
  for(llong i1=z3; i1<z4;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K_val[i] = modulus*1.1;
    }
  } 
  llong x0=120,x1=180;
  for(llong i1=z1; i1<z1+5;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*1.03;
    }
  } 
  for(llong i1=z1+5; i1<z1+10;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*0.97;
    }
  } 
  for(llong i1=z1+10; i1<z1+15;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*1.03;
    }
  } 
  for(llong i1=z1+15; i1<z1+20;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*0.97;
    }
  } 
  for(llong i1=z1+20; i1<z1+25;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*1.03;
    }
  } 
  for(llong i1=z1+25; i1<z1+30;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*0.97;
    }
  } 
}

}

#endif