#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"
#include <omp.h>
#include "base/utilities.h"

#include "base/Vector.h"
#include "base/inCoreFloat.h"
#include "base/DataInCoreFloat.h"
#include "base/Operator.h"
#include "base/MultiOperator.h"

#include "Interpolation.h"
#include "CreateModel.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Prop.h"
#include "Imaging.h"

using namespace std;
using namespace SEP;

/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/



int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "         Full Waveform inversion.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(2);   // Initialize the parallel environment
  
  // Define dimension of interest
  long long nx, nz, nt, nb;
  long long sx, sz,ns,ox,dxs;
  long long boundary;
  float dx, dz, dt;
  float velocity, density, modulus;
  float fmax=15;
  nt = 150;
  nx = 21;
  nz = 21;
  ns = 20;
  sx = 130;
  sz = 80;
  boundary = 10;
  nb = 4;

  density=2.;          // [g/cm^3]
  velocity=2.4;        // [km/s]
  dx = 10.; dz = 10.;  // [m]
  dt = 1.5;            // [ms]
  modulus=velocity*velocity*density; // [GPa]

  // Set up space 
  inCoreFloat1D sspace(nt);               // space for shot
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wave field cube
  inCoreFloat2D dspace(nt,nx);         // space for data space  

  // Set up vectors
  Vector *uf     = new Vector(&uspace);    uf->random();
  Vector *ub     = new Vector(&uspace);    ub->random();
  Vector *K      = new Vector(&mspace);    K->zeroElement();      // migration velocity
  Vector *rho    = new Vector(&mspace);    rho->zeroElement();     // Velocity vector  
  
  
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();
  K_val[10*nx+10] = 1.;
  GradZ *GX = new GradZ(&mspace,&mspace,nb);
  GX->Forward(false,K,rho);
  float *__restrict__ r_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr();
  for(llong i=0; i<nx; i++){
    cout <<"ix=" << i << " , dx = "<< r_val[7*nx+i]<<endl;
  }
  srite("DX",r_val,sizeof(float)*nx*nz);
    return 0;
  cout << "Testing interpolation module." <<endl;
  IntZ *te = new IntZ(&mspace,&mspace);
  te->DotTest(true);

  cout << "Testing SensorPad module." <<endl;   
  SensorPad *Sensor = new SensorPad(&dspace,&uspace,sz);
  Sensor->DotTest(true);
  
  cout << "Testing wave propogation." << endl;
  Models *CModel =  new Models(nz, nx, modulus, density);
  CModel->Model2(K,rho);
  Prop *Lmig = new Prop(&uspace,&uspace);    // propogation to generate data
  Lmig->SetPropPar(boundary,dt, dx, K);
  Lmig->SetModel(K,rho);
  Lmig->DotTest(true);
  
  cout << "Testing Imaging condition." << endl;
  Gradient *Imag = new Gradient(uf,K,nb,dt,dx);
  Imag->GradK(uf,ub,K);
  Imag->Gradr(uf,ub,rho);
  Imag->BornS(K,rho, uf, ub);
  cout << "Testing end." <<endl;
  return 0; 
}