#ifndef HESS_H
#define HESS_H 1
#include <cmath>
#include <vector>
#include "Taper.h"
#include "Clean.h"
#include <cmath>
#include <iostream>
#include "TimeMarch.h"
#include "base/utilities.h"
#include "Source.h"
#include "SensorPad.h"
#include "Prop.h"

typedef long long llong;
/*
Born modeling operator
Need to set velocity before running
Assume source field ready to use.

Single parameter if include Engine.h
double parameter if include EngineF.h


Created:       2015-01-15
Last modified: 2015-01-19
*/

namespace SEP {
class Hess: public Operator{
  public:
    Hess(Space *mod, Space *data, Engine *_Run, Vector *_K0, Vector *_rho0,Vector *_dr, Vector *_Hr,Vector *_Mask){
      setDomain(mod);
      setRange(data);
      Run = _Run;
      K0 = _K0;
      rho0 = _rho0;  dr = _dr; Hr = _Hr;
      Mask = _Mask;
      Ktemp1 = new Vector(mod);
      Ktemp2 = new Vector(mod);
      size=501*201;
      GN = 0;
    }
    void SetGN(int _GN){GN = _GN;}
    virtual bool Forward(bool add, Vector *model, Vector *data){
      if(!add) data->zeroElement();
      float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
      float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();  
      float *__restrict__ K1val = ((dataInCoreFloat*)Ktemp1->getDataContainer())->getFloatPtr();
      float *__restrict__ K2val = ((dataInCoreFloat*)Ktemp2->getDataContainer())->getFloatPtr();
      float *__restrict__ maskval = ((dataInCoreFloat*)Mask->getDataContainer())->getFloatPtr(); 
      for(int i=0; i < size; i++){
        K1val[i] = mval[i]*maskval[i];
      }       
      Ktemp2->zeroElement(); 
      if(GN){
        Run->GaussHessianFunc(K0,rho0,Ktemp1,dr,Ktemp2,Hr);
      }
      else{
        Run->HessianFunc(K0,rho0,Ktemp1,dr,Ktemp2,Hr);       
      }
      for(int i=0; i < size; i++){
        dval[i] += K2val[i]*maskval[i];
      }       
      //data->linearCombo(1.,model,1.);
    }
    virtual bool Adjoint(bool add, Vector *model, Vector *data){
      if(!add) model->zeroElement();
      float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
      float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();  
      float *__restrict__ K1val = ((dataInCoreFloat*)Ktemp1->getDataContainer())->getFloatPtr();
      float *__restrict__ K2val = ((dataInCoreFloat*)Ktemp2->getDataContainer())->getFloatPtr();
      float *__restrict__ maskval = ((dataInCoreFloat*)Mask->getDataContainer())->getFloatPtr();  
      for(int i=0; i < size; i++){
        K1val[i] = dval[i]*maskval[i];
      }       
      Ktemp2->zeroElement(); 
      if(GN){
        Run->GaussHessianFunc(K0,rho0,Ktemp1,dr,Ktemp2,Hr);
      }
      else{
        std::cout << "Full Newton method." <<std::endl;
        Run->HessianFunc(K0,rho0,Ktemp1,dr,Ktemp2,Hr);       
      }
      for(int i=0; i < size; i++){
        mval[i] += K2val[i]*maskval[i];
      }       
      //model->linearCombo(1.,data,1.);  
    }   
  private:
    Engine *Run;
    Vector *K0, *rho0, *dr, *Hr, *Mask;
    Vector *Ktemp1, *Ktemp2;
    int size,GN;
};


class HessAll: public Operator{
  public:
    HessAll(Space *_Kspace, Space *mod, Space *data, Engine *_Run, Vector *_K0, Vector *_rho0,Vector *_dr, Vector *_Hr){
      setDomain(mod);
      setRange(data);
      Kspace = (inCoreFloat *)_Kspace;
      mspace = (inCoreFloat *)mod;
      Ktemp1 = new Vector(Kspace);
      rtemp1 = new Vector(Kspace);
      Ktemp2 = new Vector(Kspace);
      rtemp2 = new Vector(Kspace);
      Run = _Run;
      K0 = _K0;
      rho0 = _rho0;  dr = _dr; Hr = _Hr;
      size=501*201;
    }
    virtual bool Forward(bool add, Vector *model, Vector *data){
      if(!add) data->zeroElement();
      float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
      float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();  
      float *__restrict__ K1val = ((dataInCoreFloat*)Ktemp1->getDataContainer())->getFloatPtr();
      float *__restrict__ r1val = ((dataInCoreFloat*)rtemp1->getDataContainer())->getFloatPtr(); 
      float *__restrict__ K2val = ((dataInCoreFloat*)Ktemp2->getDataContainer())->getFloatPtr();
      float *__restrict__ r2val = ((dataInCoreFloat*)rtemp2->getDataContainer())->getFloatPtr();     
      for(int i=0; i < size; i++){
        K1val[i] = mval[i];
        r1val[i] = mval[i+size];
      } 
      Ktemp2->zeroElement();
      rtemp2->zeroElement();
      Run->GaussHessianFunc(K0,rho0,Ktemp1,rtemp1,Ktemp2,rtemp2);
      for(int i=0; i < size; i++){
        dval[i] += K2val[i];
        dval[i+size] += r2val[i];
      } 
    }
    virtual bool Adjoint(bool add, Vector *model, Vector *data){
      if(!add) model->zeroElement();
      float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
      float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();  
      float *__restrict__ K1val = ((dataInCoreFloat*)Ktemp1->getDataContainer())->getFloatPtr();
      float *__restrict__ r1val = ((dataInCoreFloat*)rtemp1->getDataContainer())->getFloatPtr(); 
      float *__restrict__ K2val = ((dataInCoreFloat*)Ktemp2->getDataContainer())->getFloatPtr();
      float *__restrict__ r2val = ((dataInCoreFloat*)rtemp2->getDataContainer())->getFloatPtr();     
      for(int i=0; i < size; i++){
        K1val[i] = dval[i];
        r1val[i] = dval[i+size];
      } 
      Ktemp2->zeroElement();
      rtemp2->zeroElement();
      Run->GaussHessianFunc(K0,rho0,Ktemp1,rtemp1,Ktemp2,rtemp2);
      for(int i=0; i < size; i++){
        mval[i] += K2val[i];
        mval[i+size] += r2val[i];
      }   
    }  
 
  private:
    Space *Kspace, *mspace;
    Vector *Ktemp1, *rtemp1,*Ktemp2, *rtemp2;
    Engine *Run;
    Vector *K0, *rho0, *dr, *Hr;
    int size;
};


void ConvertFtoD(float *input, double *output, int size){
  for(int i=0; i < size; i++){
    output[i] = (double)input[i];
  }
}
void ConvertDtoF(double *input, float *output, int size){
  for(int i=0; i < size; i++){
    output[i] = (float)input[i];
  }
}

}


#endif