#ifndef INTERPOLATION_H
#define INTERPOLATION_H 1

#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "Clean.h"
#include <omp.h>
typedef long long llong;

/*
	Created by Yinbin Ma  
	Interpolation from regular grid to staggered grid
	Use carefully and have fun.
	
	First created: 2015-08-03
	Last Modified: 2015-08-03
*/	

namespace SEP {

class IntX: public Operator{
  public:
    IntX(Space *mod, Space *data){
      setDomain(mod);
      setRange(data);
      init();
    }
    ~IntX(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);

  private:
    void init();
    llong n1, n2;
    CleanVec Zero;

};

void IntX::init() {
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n1 = in_dims[0];
  n2 = in_dims[1];
}

bool IntX::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,n1,n2);
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong position;
  for(long long iz=0; iz<nz;iz++){
    for(long long ix=0; ix<nx-1;ix++){
      long long position = iz*nx+ix;
          dval[position] += (mval[position]+mval[position+1])/2;
    }  
  } 
  for(long long iz=0; iz<nz;iz++){
	  long long position = iz*nx+nx-1;
	  dval[position] += (mval[position]+mval[position])/2;
  }  
  return 0;
}

bool IntX::Adjoint(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(model,n1,n2);
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong position;
  for(long long iz=0; iz<nz;iz++){
    for(long long ix=0; ix<nx-1;ix++){
      long long position = iz*nx+ix;
          mval[position] += dval[position]/2.;
          mval[position+1] += dval[position]/2.;
    }  
  } 
  for(long long iz=0; iz<nz;iz++){
	  long long position = iz*nx+nx-1;
          mval[position] += dval[position]/2.;
          mval[position] += dval[position]/2.;
  }  
  return 0;
}


class IntZ: public Operator{
  public:
    IntZ(Space *mod, Space *data){
      setDomain(mod);
      setRange(data);
      init();
    }
    ~IntZ(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);

  private:
    void init();
    llong n1, n2;
    CleanVec Zero;

};

void IntZ::init() {
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n1 = in_dims[0];
  n2 = in_dims[1];
}

bool IntZ::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,n1,n2);
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong position;
  for(long long iz=0; iz<nz-1;iz++){
	for(long long ix=0; ix<nx;ix++){
      long long position = iz*nx+ix;
          dval[position] += (mval[position]+mval[position+nx])/2;
    }  
  } 
  for(long long ix=0; ix<nx;ix++){
    long long position = nz*(nx-1)+ix;
	  dval[position] += (mval[position]+mval[position])/2;
  }  
  return 0;
}

bool IntZ::Adjoint(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(model,n1,n2);
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong position;
  for(long long iz=0; iz<nz-1;iz++){
	for(long long ix=0; ix<nx;ix++){
      long long position = iz*nx+ix;
          mval[position] += dval[position]/2.;
          mval[position+nx] += dval[position]/2.;
    }  
  } 
  for(long long ix=0; ix<nx;ix++){
    long long position = nz*(nx-1)+ix;
	  mval[position] += dval[position]/2.;
	  mval[position] += dval[position]/2.;
  }  
  return 0;
}

}

#endif


