#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"
#include <omp.h>
#include "base/utilities.h"

#include "base/Vector.h"
#include "base/inCoreFloat.h"
#include "base/DataInCoreFloat.h"
#include "base/Operator.h"
#include "base/MultiOperator.h"

#include "Interpolation.h"
#include "CreateModel.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Prop.h"

using namespace std;
using namespace SEP;

/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/



int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "         Full Waveform inversion.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(8);   // Initialize the parallel environment
  
  // Define dimension of interest
  long long nx, nz, nt;
  long long sx, sz,ns,ox,dxs;
  long long boundary;
  float dx, dz, dt;
  float velocity, density, modulus;
  float fmax=15;
  nt = 1500;
  nx = 401;
  nz = 251;
  ns = 20;
  sx = 130;
  sz = 80;
  boundary = 10;

  density=2.;          // [g/cm^3]
  velocity=2.4;        // [km/s]
  dx = 10.; dz = 10.;  // [m]
  dt = 1.5;            // [ms]
  modulus=velocity*velocity*density; // [GPa]

  // Set up space 
  inCoreFloat1D sspace(nt);               // space for shot
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wave field cube
  inCoreFloat3D dspace(ns,nt,nx);         // space for data space
  // Set up vectors
  Vector *K      = new Vector(&mspace);    K->zeroElement();      // migration velocity
  Vector *rho    = new Vector(&mspace);    rho->zeroElement();     // Velocity vector
  Vector *K0      = new Vector(&mspace);    K0->zeroElement();      // migration velocity
  Vector *rho0    = new Vector(&mspace);    rho0->zeroElement();     // Velocity vector
  Vector *dK      = new Vector(&mspace);    dK->zeroElement();      // migration velocity
  Vector *drho    = new Vector(&mspace);    drho->zeroElement();     // Velocity vector
  Vector *GradK  = new Vector(&mspace);    GradK->zeroElement();      // migration velocity
  Vector *Gradr  = new Vector(&mspace);    Gradr->zeroElement();     // Velocity vector
  Vector *vel    = new Vector(&mspace);    vel->zeroElement();      // migration velocity
  Vector *shot   = new Vector(&sspace);    shot->zeroElement();     // shot profile
  Vector *rec    = new Vector(&sspace);    rec->zeroElement();     // shot profile
  Vector *rec2   = new Vector(&sspace);    rec2->zeroElement();     // shot profile
  Vector *rec0   = new Vector(&sspace);    rec0->zeroElement();     // shot profile
  Vector *data   = new Vector(&dspace);    data->zeroElement();     // data
  Vector *us     = new Vector(&uspace);    us->zeroElement();       // Wave field
  Vector *u0     = new Vector(&uspace);    u0->zeroElement();       // Wave field
  Vector *u1     = new Vector(&uspace);    u1->zeroElement();       // Wave field
  Vector *v1     = new Vector(&uspace);    v1->zeroElement();       // Wave field  
  Vector *v2a     = new Vector(&uspace);    v2a->zeroElement();       // Wave field 
  Vector *v2b    = new Vector(&uspace);    v2b->zeroElement(); 
  Vector *utemp1 = new Vector(&uspace);    utemp1->zeroElement();       // Wave field
  Vector *GNHessK  = new Vector(&mspace);    GNHessK->zeroElement();      // migration velocity  
  Vector *HessK1  = new Vector(&mspace);    HessK1->zeroElement();      // migration velocity
  Vector *HessK2  = new Vector(&mspace);    HessK2->zeroElement();      // migration velocity
  float *__restrict__ rec_val = ((dataInCoreFloat*)rec->getDataContainer())->getFloatPtr(); 
  float *__restrict__ u0_val = ((dataInCoreFloat*)u0->getDataContainer())->getFloatPtr();   
  float *__restrict__ u1_val = ((dataInCoreFloat*)u1->getDataContainer())->getFloatPtr();  
  float *__restrict__ v1_val = ((dataInCoreFloat*)v1->getDataContainer())->getFloatPtr();
  float *__restrict__ v2a_val = ((dataInCoreFloat*)v2a->getDataContainer())->getFloatPtr();
  float *__restrict__ v2b_val = ((dataInCoreFloat*)v2b->getDataContainer())->getFloatPtr();
  float *__restrict__ utemp_val = ((dataInCoreFloat*)utemp1->getDataContainer())->getFloatPtr();
  float *__restrict__ us_val = ((dataInCoreFloat*)us->getDataContainer())->getFloatPtr();
  // Read in source wavelet and velocity field
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  sreed("wavelet",shot_val,sizeof(float)*nt); 

  // Set up Inter
  IntZ *te = new IntZ(&mspace,&mspace);
  te->DotTest(true);
  return 0;

  // Set up model
  Models *CModel =  new Models(nz, nx, modulus, density);
  CModel->Model6(K,rho);
  CModel->Model2(K0,rho0);
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();
  float *__restrict__ K0_val = ((dataInCoreFloat*)K0->getDataContainer())->getFloatPtr(); 
  float *__restrict__ dK_val = ((dataInCoreFloat*)dK->getDataContainer())->getFloatPtr(); 
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr(); 
  srite("modelK",K_val,sizeof(float)*nx*nz); 
  srite("modelr",rho_val,sizeof(float)*nx*nz); 
  //return 0;
  // Check numerical condition
  cout << "    Dispersion: Vmin/(fmax*dx) = " << 1000.*velocity/fmax/dx <<", should be > 4"<<endl;
  cout << "    Courant:    Vmax*dt/dx     = " << velocity*dt/dx <<", should be < 0.717"<<endl;
 
  // Generate operators for wave progation
  Prop *Lmig = new Prop(&uspace,&uspace);    // propogation to generate data
  Prop *Lmig0= new Prop(&uspace,&uspace);    // propagation using migration velocity
  Lmig->SetPropPar(boundary,dt, dx, K);
  Lmig0->SetPropPar(boundary,dt, dx, K);
  Lmig->SetModel(K,rho);
  Lmig0->SetModel(K0,rho0);
  //float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr(); 
  srite("truev",K_val,sizeof(float)*nz*nx);
  srite("migv",K0_val,sizeof(float)*nz*nx);
  K->linearCombo(1.,*K0,-1.); dK->linearCombo(0.,*K,1.);
  rho->linearCombo(1.,*rho0,-1.); drho->linearCombo(0.,*rho,1.);
  srite("deltav",dK_val,sizeof(float)*nz*nx);
  // Inject source wavelet
  Source *Ss = new Source(nt, nz, nx);
  //Ss->Injection(shot, u1, sz, sx);
 
  SensorPad *SP = new SensorPad(&dspace, &uspace, sz, boundary);
  SP->SetShot(0);  
  for(llong is=0; is < ns; is++){
    cout << "Working on shot " << is << " out of " << ns << endl;
    sx = 100+is*10;
    Ss->Injection(shot, us, sz, sx);
    // Forward propagation for u field and data
    double t0 = timer();
    Lmig0->Forward(false,us,u0);           // u is computed 
    Lmig->Forward(false,us,utemp1);         // source -> ufield
	utemp1->linearCombo(1.,*u0,-1.);
    SP->Adjoint(false,data,utemp1);  
    double t1 = timer(); 
    cout << "   wave propagation, takes " << t1-t0 << "seconds. " <<endl;
   
    // Now Adjoint propagation for lambda
    t0 = timer();
    SP->Forward(false,data,utemp1);      
    Lmig0->Adjoint(false,v1,utemp1);   // v1 = lambda
    t1 = timer(); 
    cout << "   lambda wave propagation, takes " << t1-t0 << "seconds. " <<endl;
   // srite("bwave1",v1_val,sizeof(float)*nz*nx*nt); 
    
    // Now for delta_u_1
    t0 = timer();  
    Lmig0->BornS(dK,drho,u0,us);    // now us contains [*****]u
    Lmig0->Forward(false,us,u1);    // u1=delta_u1
    t1 = timer(); 
    cout << "   delta_u1 wave propagation, takes " << t1-t0 << "seconds. " <<endl;  

    // Now for mu_b
    t0 = timer(); 
    us->zeroElement();
    Lmig0->BornS(dK,drho,v1,us);    // [****]lambda
    //SP->Forward(true,data,us);      // SS delta_u + [*****]lambda
    Lmig0->Adjoint(false,v2b,us);
    t1 = timer(); 
    cout << "   mu wave propagation, takes " << t1-t0 << "seconds. " <<endl;
  
    // Now for mu_a
    t0 = timer(); 
    us->zeroElement();
    SP->Adjoint(false,data,u1); 
    SP->Forward(false,data,us); 
    //SP->Forward(true,data,us);      // SS delta_u + [*****]lambda
    Lmig0->Adjoint(false,v2a,us);
    t1 = timer(); 
    cout << "   eta wave propagation, takes " << t1-t0 << "seconds. " <<endl; 
    
    Lmig0->kernel_K(u0,v1,GradK);
    Lmig0->kernel_K(u0,v2a,GNHessK);
    Lmig0->kernel_K(u0,v2b,HessK2);
    Lmig0->kernel_K(u1,v1,HessK1);
  } 
  // Now let us compute kernel
  float *__restrict__ gk_val = ((dataInCoreFloat*)GradK->getDataContainer())->getFloatPtr();  
  float *__restrict__ gr_val = ((dataInCoreFloat*)Gradr->getDataContainer())->getFloatPtr();

  srite("GradKR",gk_val,sizeof(float)*nz*nx);
  srite("GradrR",gr_val,sizeof(float)*nz*nx); 
  
  float *__restrict__ GNHessK_val = ((dataInCoreFloat*)GNHessK->getDataContainer())->getFloatPtr();  
  srite("GNHessKR",GNHessK_val,sizeof(float)*nz*nx);
  
  float *__restrict__ HessK2_val = ((dataInCoreFloat*)HessK2->getDataContainer())->getFloatPtr(); 
  srite("HessKR2",gk_val,sizeof(float)*nz*nx);

  float *__restrict__ HessK1_val = ((dataInCoreFloat*)HessK1->getDataContainer())->getFloatPtr(); 
  srite("HessKR1",gk_val,sizeof(float)*nz*nx);

  return 0;  
  // output wavefield




    
  return 0;
  /*
  float *__restrict__ gk_val = ((dataInCoreFloat*)GradK->getDataContainer())->getFloatPtr();  
  float *__restrict__ gr_val = ((dataInCoreFloat*)Gradr->getDataContainer())->getFloatPtr();
  float *__restrict__ rec_val = ((dataInCoreFloat*)rec->getDataContainer())->getFloatPtr();
  float *__restrict__ rec2_val = ((dataInCoreFloat*)rec2->getDataContainer())->getFloatPtr();
   float *__restrict__ rec0_val = ((dataInCoreFloat*)rec0->getDataContainer())->getFloatPtr();   
 
  for(llong i=0; i<nt;i++){
      rec_val[i] = (u3_val[i*nx*nz+sz*nx+nx-sx]-u2_val[i*nx*nz+sz*nx+nx-sx]); //-(u3_val[i*nx*nz+(sz+1)*nx+nx-sx]-u2_val[i*nx*nz+(sz+1)*nx+nx-sx]);
      rec2_val[i] = u3_val[i*nx*nz+sz*nx+nx-sx]; //-u3_val[i*nx*nz+(sz+1)*nx+nx-sx];
      rec0_val[i] = u2_val[i*nx*nz+sz*nx+nx-sx];
   }    
  srite("rec1",rec_val,sizeof(float)*nt); 
  srite("rec2",rec2_val,sizeof(float)*nt); 

  u2->zeroElement();
  //sreed("low",rec_val,sizeof(float)*nt);
  for(llong i=0; i<nt;i++){
    u2_val[i*nx*nz+sz*nx+nx-sx] = rec_val[i];
    //u2_val[i*nx*nz+(sz+1)*nx+nx-sx] = -1.*rec_val[i];
   }
  //u3->addInv();  
  Lmig->Forward(false,u1,u3);
  Lmig->Adjoint(false,u1,u2);  
  Lmig->kernel_K(u3,u1,GradK);
  Lmig->kernel_r(u3,u1,Gradr);
  srite("GradKT",gk_val,sizeof(float)*nz*nx);
  srite("GradrT",gr_val,sizeof(float)*nz*nx);
  u1->zeroElement();
  Ss->Injection(shot, u1, sz, sx);
  Lmig0->Forward(false,u1,u3);
  Lmig0->Adjoint(false,u1,u2);
  GradK->addInv();
  Gradr->addInv();
  Lmig->kernel_K(u3,u1,GradK);
  Lmig->kernel_r(u3,u1,Gradr);

  //Lmig->kernel_r(u3,u1,Gradr);
  u2->zeroElement();
  
  
  //Lmig->kernel_bigr(u3,u1,u2);
  srite("GradKR",gk_val,sizeof(float)*nz*nx);
  srite("GradrR",gr_val,sizeof(float)*nz*nx);   
  srite("wave1",u3_val,sizeof(float)*nz*nx*nt); 
  srite("wave2",u1_val,sizeof(float)*nz*nx*nt); 
  srite("buildup",u2_val,sizeof(float)*nz*nx*nt);

  u2->zeroElement();
  //sreed("low",rec_val,sizeof(float)*nt);
  for(llong i=0; i<nt;i++){
    u2_val[i*nx*nz+sz*nx+nx-sx] = rec2_val[i];
   } 
  u1->zeroElement();
  Ss->Injection(shot, u1, sz, sx);
  Lmig->Forward(false,u1,u3); 
  Lmig->Adjoint(false,u1,u2);  
  GradK->zeroElement();
  Gradr->zeroElement();
  Lmig->kernel_K(u3,u1,GradK);
  Lmig->kernel_r(u3,u1,Gradr);
  srite("GradKT",gk_val,sizeof(float)*nz*nx);
  srite("GradrT",gr_val,sizeof(float)*nz*nx);
  
  return 0;
  */
   /*
  // Save data
  SensorPad *SP = new SensorPad(&dspace, &uspace, sz, boundary);
  SP->SetShot(0);  
  SP->Adjoint(false,data,u3);
  
  // Output data  
  float *__restrict__ u3_val = ((dataInCoreFloat*)u3->getDataContainer())->getFloatPtr();
  float *__restrict__ d_val = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr(); 
  srite("wave1",u3_val,sizeof(float)*nz*nx*nt);
  srite("data1",d_val,sizeof(float)*nx*nt); 

  // Second model
  CModel->Model1(K,rho);
  Lmig->SetModel(K,rho);
  Lmig->Forward(false,u1,u3);
  SP->Adjoint(false,data,u3);
  srite("wave2",u3_val,sizeof(float)*nz*nx*nt);
  srite("data2",d_val,sizeof(float)*nx*nt);    
  // Clean up
*/
  
   std::cout << "========== Done, Cleaning up ==========" <<std::endl; 

  return 0;
}