#ifndef NCG_H
#define NCG_H 1

#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "base/Algorithm.h"
#include "Clean.h"
#include "Func.h"
#include <omp.h>
typedef long long llong;

/*
	Created by Yinbin Ma  
	4th order Laplacian in 2 d.
	Use carefully and have fun.
	
	First created: 2015-01-01
	Last Modified: 2015-02-28
*/	

namespace SEP {
class Zoom{
  public:
    Zoom(Space *space){
      mspace = (inCoreFloat *)space;
      xtr = new Vector(mspace);
      gradtr = new Vector(mspace);
    }  

    int run(double *alpha, Vector *m, double *f, Vector *grad, int *nsteps,
             double low, double high, double flow, double fhigh, double glow, double ghi, 
             double f0, double g0, Vector *m0, Vector *d,
             double c1, double c2,FuncGrad *Func){
      // input: 
      // output: alpha, m, f, grad, nsteps
      fail=0;
      //*nsteps = 0; 
      double alphatemp = low;
      while(fabs(low-high) >1.e-3 && *nsteps <100){
        std::cout <<"===========ZOOM==============="<<std::endl;
      //std::cout <<"low = "<<low<<", high="<<high<<std::endl;
        alphatr = (low+high)/2.;    // bisection
        xtr->linearCombo(0.,*m0,1.);
        xtr->linearCombo(1.,*d,(float)alphatr);
        Func->Evaluate(xtr,&ftr,gradtr);*nsteps = *nsteps+1;
        gtr=gradtr->innerProd(*d);
        std::cout <<"alphatr=" << alphatr<<std::endl;
        std::cout <<"dd" << gradtr->innerProd(*gradtr) << std::endl;
        std::cout <<"gtr = " << gtr <<", g0 = " << g0 << std::endl;
        std::cout <<"ftr = " << ftr <<", f0 = " << f0 <<",flow=" << flow<< std::endl;
        if(ftr > f0 + c1*alphatr*g0 || ftr >= flow){
          // overshot
          //std::cout <<"        over shot"<<std::endl;
          high=alphatr; fhigh = ftr;
        }
        else {
          if(fabs(gtr) <= -c2*g0){
            *alpha = alphatr; m->linearCombo(0.,*xtr,1.);
            *f=ftr; grad->linearCombo(0.,*gradtr,1.);
            std::cout << "    Zoom successful, f= "<<*f << std::endl;
            return 0;
          }
          if(gtr*(high-low) >=0){
            high = low; fhigh = flow;
          }
          low=alphatr; flow=ftr;
        }
        std::cout <<"low = " << low <<", high = " << high<< std::endl;
      }
      // failed

      *alpha = alphatr; m->linearCombo(0.,*xtr,1.);
      *f=ftr; grad->linearCombo(0.,*gradtr,1.);
      std::cout << "    Zoom failed,"<<*alpha <<",f ="<<*f << std::endl;
      return 1;
    }
   /*
       int run(Vector *m, Vector *m0, FuncGrad *Func){
      m->linearCombo(0.,m0,1.);
      Func->Evaluate(m,&f,grad);

      gnorm = sqrt(grad->innerProd(*grad));
      std::cout << "starting from f="<< f<<", gnorm = "<<gnorm << std::endl;
      if(gnorm < normtol) {
        std::cout << "ncg: grad norm < tolorance at initial iteration\n" << std::endl;
        return 0;
      }
      d->linearCombo(0.,grad,-1.);   // p=-g for steepest descent
      for(int iter=0; iter < niter; iter++){
        gtp = grad->innerProd(*d);
        if(gtp >=0 ){ 
          std::cout << "NCG: not a descent direction, restart linesearch" << std::endl;
          Func->Evaluate(m,&f,grad);
          d->linearCombo(0.,grad,-1.); 
        }
        // Line search 
        mprev->linearCombo(0.,*m,1.);
        gprev->linearCombo(0.,*grad,1.);
        fprev = f;
        //Func->Evaluate(m,&fprev,gprev);
        lsch->run(&alpha, m, &f, grad, &nsteps, mprev,fprev,gprev, d, Func); // Line search
        lsch->SetPar(1.5*alpha);
        gnorm = sqrt(grad->innerProd(*grad));
        std::cout << "NCG iteration: "<< iter << ", step = " << alpha << ", f = "<< f <<", gnorm = "<<gnorm << ", nsteps = " << nsteps<< std::endl;
        if(gnorm <normtol){
          std::cout << " accuracy reached." << std::endl;
          return 0;
        }
        //lsch->SetPar(alpha);
        // Assume everything is fine
        y->linearCombo(0.,*grad,1.);
        y->linearCombo(1.,*gprev,-1.);
        //y = grad-gprev;
        // PR+
        nmgprevsq = gprev->innerProd(*gprev);
        beta_pr = grad->innerProd(*y)/nmgprevsq;
        beta_fr = grad->innerProd(*grad)/nmgprevsq;
        if(beta_pr < -beta_fr){
          beta=-beta_fr;
        }
        else if (beta_pr > beta_fr){
          beta = beta_fr;
        }
        else{
          beta = beta_pr;
        }
        d->linearCombo(beta,*grad,-1.);
      }
      std::cout << " max number of iteration reached." << std::endl;
    }
    */
   
  private:
    inCoreFloat *mspace;    
    int fail, nsteps,in_int;
    double low2, flow2, glow2, bisect, eta, gamma,sgn,interp,alphatr,ftr,gtr;
    Vector *xtr, *gradtr;
     //   Func->Evaluate(mnew,&fnew,gradnew);
};



class LineSearch{
  public:
    LineSearch(Space *space){
      mspace = (inCoreFloat *)space;
      zoom_in = new Zoom(mspace);
      c1 = 0;
      c2 = 0.9;
      niter = 50;
      mnew = new Vector(mspace);
      gradnew = new Vector(mspace);
      alphamax=5.;
    }
    void SetPar(double _alphamax){alphamax = _alphamax;}

    int run(double *alpha, Vector *m, double *f, Vector *grad, int *nsteps,
             Vector *m0, double f0, Vector *grad0, Vector *d, FuncGrad *Func){
      g0 = grad0->innerProd(*d);  
      std::cout << "g0 = " << g0 << std::endl;
      if(g0 >=0){
        std::cout << "Line search: not a descent direction!" <<std::endl;
      }
      old = 0; fold=f0; gold=g0;
      alphanew =alphamax;
      for(int iter=0; iter < niter; iter++){
        std::cout << "    Line iteration: " << iter << std::endl;
        mnew->linearCombo(0.,*m0,1.);
        mnew->linearCombo(1.,*d,(float)alphanew);
        Func->Evaluate(mnew,&fnew,gradnew);  *nsteps = *nsteps+1;
        //std::cout << "norm m = " << mnew->innerProd(*mnew) << std::endl;
        std::cout << "alphanew=" << alphanew<<", fnew = " << fnew << std::endl;
        gnew = gradnew->innerProd(*d);
        //std::cout <<"gg = " << gradnew->innerProd(*gradnew)<<"gnew = " << gnew <<", g0 = " << g0 << std::endl;
        if(fnew > f0 + c1*alphanew*g0 || ((fnew >= fold)&&iter>0)){
          std::cout << " zoom in norm order:" << old <<"to "<< alphanew<< std::endl;
          zoom_in->run(alpha,m,f,grad,nsteps,
                       old,alphanew,fold,fnew,gold,gnew,
                       f0,g0,m0,d,c1,c2,Func);
          std::cout << "    Haha, Zoom successful, f= "<<*f << std::endl;
          //*alpha = alphanew;
          //std::cout << "mnorm" << m->innerProd(*m) <<std::endl;
          //std::cout << " Line search successful. case 1" << std::endl;
          return 0;
        }
        if(fabs(gnew) <= -c2*g0){
          *alpha = alphanew; m->linearCombo(0.,*mnew,1.); 
          *f = fnew; grad->linearCombo(0.,*gradnew,1.);
          return 0;
        }
        if(gnew >=0){
          std::cout << " zoom in norm order:" << alphanew <<"to "<< old<< std::endl;
           zoom_in->run(alpha,m,f,grad,nsteps,
                       alphanew,old,fnew,fold,gnew,gold,
                       f0,g0,m0,d,c1,c2,Func);
          std::cout << "    Haha, Zoom successful, f= "<<*f << std::endl;
          //*alpha = alphanew;
          //std::cout << " Line search successful. case 2" << std::endl;
          return 0;      
        }
        old = alphanew;
        fold = fnew;
        gold = gnew;
        alphanew = 2*alphanew;
      }  // end of for
      std::cout <<"line search: minimizer not bracketed" << std::endl;
      *alpha = alphanew; 
      m->linearCombo(0.,*mnew,1.); 
      *f = fnew; grad->linearCombo(0.,*gradnew,1.);
    }  
    ~LineSearch(){}
  
  private:
    inCoreFloat *mspace;
    Zoom *zoom_in;
    Vector *mnew,*gradnew;
    double g0, old,gold;
    double c1; // Wolfe condition
    double c2; // curvature condition
    double alphamax;
    int niter;
    double fold, fnew, alphanew,gnew;
    
};

class NCG{
  public:

    NCG(Space *space){
      mspace = (inCoreFloat *)space;
      lsch = new LineSearch(mspace);
      grad = new Vector(mspace);
      gprev = new Vector(mspace);
      y = new Vector(mspace);
      d = new Vector(mspace);
      mprev = new Vector(mspace);
      niter = 50;
      normtol = 1.e-5;
      mblock=501*201;
    }
    void SetPar(int _niter){niter = _niter;}
    int run(Vector *m, Vector *m0, FuncGrad *Func, int *nfunc, float *objs,float *Ks,float *GKs,float *dKs){
      float *__restrict__ m_val = ((dataInCoreFloat*)m->getDataContainer())->getFloatPtr();
      float *__restrict__ d_val = ((dataInCoreFloat*)d->getDataContainer())->getFloatPtr();
      float *__restrict__ g_val = ((dataInCoreFloat*)grad->getDataContainer())->getFloatPtr();
      m->linearCombo(0.,m0,1.);
      Func->Evaluate(m,&f,grad);

      gnorm = sqrt(grad->innerProd(*grad));
      std::cout << "starting from f="<< f<<", gnorm = "<<gnorm << std::endl;
      if(gnorm < normtol) {
        std::cout << "ncg: grad norm < tolorance at initial iteration\n" << std::endl;
        return 0;
      }
      d->linearCombo(0.,grad,-1.);   // p=-g for steepest descent
      std::cout << "Max iter = " << niter << std::endl;
      for(int iter=0; iter < niter; iter++){
        nsteps = 0;
        gtp = grad->innerProd(*d);
        if(gtp >=0 ){ 
          std::cout << "NCG: not a descent direction, restart linesearch" << std::endl;
          Func->Evaluate(m,&f,grad);
          d->linearCombo(0.,grad,-1.); 
        }
        // Line search 
        mprev->linearCombo(0.,*m,1.);
        gprev->linearCombo(0.,*grad,1.);
        fprev = f;
        //Func->Evaluate(m,&fprev,gprev);
        lsch->run(&alpha, m, &f, grad, &nsteps, mprev,fprev,gprev, d, Func); // Line search
        lsch->SetPar(1.3*alpha);
        gnorm = sqrt(grad->innerProd(*grad));
        objs[iter+1]=f;
        nfunc[iter+1] = nfunc[iter] + nsteps;

        std::cout << "NCG iteration: "<< iter << ", step = " << alpha << ", f = "<< f <<", gnorm = "<<gnorm << ", nsteps = " << nfunc[iter+1]<< std::endl;

        if(gnorm <normtol){
          std::cout << " accuracy reached." << std::endl;
          return 0;
        }
        //lsch->SetPar(alpha);
        // Assume everything is fine
        y->linearCombo(0.,*grad,1.);
        y->linearCombo(1.,*gprev,-1.);
        //y = grad-gprev;
        // PR+
        nmgprevsq = gprev->innerProd(*gprev);
        beta_pr = grad->innerProd(*y)/nmgprevsq;
        beta_fr = grad->innerProd(*grad)/nmgprevsq;
        if(beta_pr < -beta_fr){
          beta=-beta_fr;
        }
        else if (beta_pr > beta_fr){
          beta = beta_fr;
        }
        else{
          beta = beta_pr;
        }
        d->linearCombo(beta,*grad,-1.);
        std::cout << "beta  = " << std::endl;
      for(int j=0;j<mblock;j++){
        Ks[iter*mblock+j] = m_val[j];
        dKs[iter*mblock+j] = d_val[j];
        GKs[iter*mblock+j] = g_val[j];
        }
      }

      std::cout << " max number of iteration reached." << std::endl;
    }
    void GetLast(Vector *DK, Vector *GK){
      DK->linearCombo(0.,*d,1.);
      GK->linearCombo(0.,*grad,1.);
    }
    ~NCG() {}

  private:
    // reference to external objects
    Vector *y, *grad, *gprev,*mprev,*d;
    inCoreFloat *mspace;
    LineSearch *lsch;
    double beta, beta_fr, beta_pr, gtp,nmgprevsq,alpha,f,fprev, gnorm;
    double normtol;
    int niter, nsteps;
    int mblock;
    
};


}
#endif
