#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"

#include <omp.h>
#include "base/utilities.h"

#include "base/s_reg.h"
#include "base/hypercube.h"
#include "base/Vector.h"
#include "base/inCoreFloat.h"
#include "base/DataInCoreFloat.h"
#include "base/Operator.h"
#include "base/MultiOperator.h"
#include "base/CGAlg.h"
#include "base/matmul.h"
#include "Interpolation.h"
#include "CreateModel.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Prop.h"
#include "Imaging.h"
#include "Engine.h"
#include "NCG.h"
#include "Hess.h"
#include "Rand.h"

#include "mkl_rci.h"
#include "mkl_blas.h"
#include "mkl_spblas.h"
#include "mkl_service.h"


extern "C" {
#include "sep3d.h"
}
using namespace std;
using namespace SEP;

/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/

extern "C"  int sep_get_data_axis_par(const char *r,int *ax,int *ff,float *o,float *d,char *l);

/*
float ObjFuc(float *dataRes, int len){
  float obj=0.;
  for(int i=0; i < len; i++){
    obj += dataRes[i]*dataRes[i];
  }
  return obj;
}

*/

int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "         Full Waveform inversion.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(10);   // Initialize the parallel environment
  
  // Define dimension of interest
  int nx, nz, nt;
  int init,flag;
  int sx, sz,ns,oxs,dxs;
  int boundary;
  float dx, dz, dt;
  float velocity, density, modulus;
  float fmax=15;
  int nb, bsize;
  int GN;
 
   // Get parameters
  getch("nb","d",&nb);    cout << "	nb = " << nb << endl;
  getch("bsize","d",&bsize);    cout << "	bsize = " << bsize << endl;
  getch("nx","d",&nx);    cout << "	nx = " << nx << endl;
  getch("nz","d",&nz);    cout << "	nz = " << nz << endl;
  getch("nt","d",&nt);   cout << "	nt = " << nt << endl;
  getch("boundary","d",&boundary);    cout << "	boundary = " << boundary << endl;
  getch("dx","f",&dx);    cout << "	dx = " << dx <<"[m]"<< endl;
  dz = dx;
  getch("sz","d",&sz);   cout << "	sz = " << sz << endl;
  getch("oxs","d",&oxs);  cout << "	oxs = " << oxs << endl;
  getch("dxs","d",&dxs);  cout << "	dxs = " << dxs << endl;
  getch("dt","f",&dt);    cout << "	dt = " << dt << "[ms]"<< endl;
  getch("ns","d",&ns);     cout << "	ns = " << ns << endl;
  getch("density","f",&density); cout << "	density = " << density << "[g/cm^3]" <<endl;
  getch("velocity","f",&velocity); cout << "	velocity = " << velocity << "[km/s]" <<endl;
  getch("init","d",&init);    cout << "	init = " << init << endl;
  getch("GN","d",&GN);    cout << "	GN = " << GN << endl;
  getch("Newton","d",&flag);    cout << "	flag = " << flag << endl;
  modulus=velocity*velocity*density;  cout << "	modulus = " << modulus << "[GPa]" <<endl;
   
  //nt =400;
  // Set up space 
  inCoreFloat1D sspace(nt);               // space for shot
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wave field cube
  inCoreFloat2D dspace(nt,nx);            // space for data space

  // Set up vectors
  Vector *K      = new Vector(&mspace);    K->zeroElement();      // migration velocity
  Vector *Mask      = new Vector(&mspace);    Mask->zeroElement();      // migration velocity
  Vector *Maskout    = new Vector(&mspace);    Maskout->zeroElement();      // migration velocity
  Vector *Ktmp      = new Vector(&mspace);    Ktmp->zeroElement();      // migration velocity
  Vector *Ktmp2      = new Vector(&mspace);    Ktmp2->zeroElement();      // migration velocity
  Vector *rho    = new Vector(&mspace);    rho->zeroElement();     // Velocity vector
  Vector *dK      = new Vector(&mspace);    dK->zeroElement();      // migration velocity
  Vector *drho    = new Vector(&mspace);    drho->zeroElement();     // Velocity vector
  Vector *K0      = new Vector(&mspace);    K0->zeroElement();      // migration velocity
  Vector *rho0    = new Vector(&mspace);    rho0->zeroElement();     // Velocity vector
  Vector *HK      = new Vector(&mspace);    HK->zeroElement();      // migration velocity
  Vector *Hrho    = new Vector(&mspace);    Hrho->zeroElement();     // Velocity vector
  Vector *shot   = new Vector(&sspace);    shot->zeroElement();     // shot profile

  // Read in wavelet and construct model
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  sreed("wavelet",shot_val,sizeof(float)*nt); 

  // Set up source 
  std::vector<int> sxv;
  std::vector<int> szv;
  for(int is=0; is < ns; is++){
    sxv.push_back(oxs+is*dxs);
    //sxv.push_back(150);
    szv.push_back(sz);
  }
  // Set up model
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();
  float *__restrict__ Ktmp_val = ((dataInCoreFloat*)Ktmp->getDataContainer())->getFloatPtr();
  float *__restrict__ Ktmp2_val = ((dataInCoreFloat*)Ktmp2->getDataContainer())->getFloatPtr();
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr(); 
  float *__restrict__ K0_val = ((dataInCoreFloat*)K0->getDataContainer())->getFloatPtr();
  float *__restrict__ rho0_val = ((dataInCoreFloat*)rho0->getDataContainer())->getFloatPtr();
  float *__restrict__ dK_val = ((dataInCoreFloat*)dK->getDataContainer())->getFloatPtr();
  float *__restrict__ drho_val = ((dataInCoreFloat*)drho->getDataContainer())->getFloatPtr();  
  float *__restrict__ HK_val = ((dataInCoreFloat*)HK->getDataContainer())->getFloatPtr();
  float *__restrict__ Hrho_val = ((dataInCoreFloat*)Hrho->getDataContainer())->getFloatPtr();   
  float *__restrict__ mask_val = ((dataInCoreFloat*)Mask->getDataContainer())->getFloatPtr();
  float *__restrict__ maskout_val = ((dataInCoreFloat*)Maskout->getDataContainer())->getFloatPtr();
  // Set up Engine
  Engine *Run = new Engine(&mspace,&dspace,&uspace,shot,sxv,szv);
  Run->SetPropPar(nb, bsize, boundary, dt, dx, dz); 
  float *dataTrue=new float[ns*nx*nt];
  float *dataDirect=new float[ns*nx*nt]; 
  float *dataRes=new float[ns*nx*nt]; 
    
  Models *CModel =  new Models(nz, nx, modulus, density);
  CModel->Model9(K,rho);
  CModel->Model8(K0,rho0); 
  cout <<"here" <<endl;
  sreed("modelK0",K0_val, sizeof(float)*nx*nz);
  sreed("modelK",K_val, sizeof(float)*nx*nz);

  cout <<"there" <<endl;
  if(flag==0){
  cout << "Need to compute direction!" <<endl;
  // Collect data residual and inject data residual
  cout << "start to run " <<endl;
  Run->SetModel(K,rho);
  Run->CreateData();
  Run->GetData(dataTrue);
  //sreed("modelK0",K0_val,sizeof(float)*nz*nx);
  Run->SetModel(K0,rho0);
  Run->CreateData();
  Run->GetData(dataRes);

  //for(int i=0; i < nt*nx*ns; i++){ dataRes[i] = 0;}
  //for(int i=0; i < nt; i++){ dataRes[i*nx+nx-sx] = dataRes[i*nx+nx-sx]-dataTrue[i*nx+nx-sx];}
  for(int j=0; j < nt*nx*ns; j++) dataRes[j] = dataRes[j]-dataTrue[j];
  float obj0 = ObjFuc(dataRes,nt*nx*ns);
  cout << "Initial model obj = " << obj0 << endl;  
  
  // Hessian Test
  //dK->linearCombo(0.,K,1.); dK->linearCombo(1.,K0,-1.);
  //Run->HessianFunc(K0,rho0,dK,drho,HK,Hrho);
  Run->PutData(dataRes);
  Run->GradientFunc(HK,Hrho,K0,rho0);  HK->addInv();
  RandW(Mask, Maskout, 70);
  for(int i=0;i<nx*nz;i++)HK_val[i] = HK_val[i]*mask_val[i];
  srite("GradK",HK_val,sizeof(float)*nz*nx);
  //return 0;
  cout << "Initial HK obj = " << sqrt(HK->innerProd(*HK)) << endl;
  Hess *HessOP = new Hess(&mspace,&mspace,Run, K0,rho0,drho,Hrho,Maskout);

  HessOP->SetGN(GN);

  //  dK->linearCombo(0.,*HK,1.);
  //  HessOP->Adjoint(false,HK,dK);
  //HessOP->DotTest(true);
  //return 0;

  
  // Set up MKL
  double rtol;
  float rtemp;
  getch("rtol","f",&rtemp);rtol=rtemp;    cout << "	rtol = " << rtol << endl;
  MKL_INT n=nx*nz, rci_request, itercount, expected_itercount=100,i;
  MKL_INT length = 128;
  //double *expected_sol = new double[n];
  //ConvertFtoD(x_val,expected_sol,n);
  double *rhs = new double[n];
  float *resid = new float[nx*nz];
  ConvertFtoD(HK_val,rhs,n);
  double *solution = new double[n]; //for(int i=0; i<n;i++) solution[i]=0.;
  ConvertDtoF(solution,dK_val,n);
  cout << "0 = " << dK->innerProd(*dK) <<endl;
  double *tmp = new double[n*4]; for(int i=0; i<4*n;i++) tmp[i]=0.;
  double *dpar= new double[128];
  MKL_INT *ipar = new MKL_INT[128];
  dcg_init(&n,solution,rhs,&rci_request,ipar,dpar,tmp);
  ipar[8] = 1;
  ipar[9] = 0;
  dpar[0] = rtol;
  dcg_check(&n,solution,rhs,&rci_request,ipar,dpar,tmp);


  double t0 = timer();
  for(int iter = 0 ; iter < 10; iter++){
    cout<<"iter = " << iter <<endl;
    
    dcg(&n,solution,rhs,&rci_request,ipar,dpar,tmp);
    ConvertDtoF(tmp,dK_val,n);
    if(rci_request == 0){
      break;
    }
    HessOP->Forward(false,dK,Ktmp);
    //mul->Forward(false,xtmp,btmp);
    ConvertFtoD(Ktmp_val,&tmp[n],n);
    
  }
  double t1 = timer();
  dcg_get(&n,solution,rhs,&rci_request,ipar,dpar,tmp,&itercount);
  ConvertDtoF(solution,dK_val,n);

  cout << "Niter = " << 5<< ", time = " << t1-t0 << endl;
  //HK->addInv();
  //HessOP->Forward(true,dK,HK);
  //cout <<" residue " << sqrt(HK->innerProd(*HK)) << endl;
  ConvertDtoF(&tmp[2*n],HK_val,n);
  std::cout <<"residue = " <<sqrt(HK->innerProd(*HK))  << endl;
  //for(int i=0;i<nx*nz;i++)HK_val[i] = HK_val[i]*mask_val[i];
  srite("HessK",dK_val,sizeof(float)*nz*nx);
  

  }  

/* 
  // Set up 
  inCoreFloat1D xspace(20);
  inCoreFloat1D bspace(20);
  inCoreFloat2D aspace(20,20);
  Vector *x = new Vector(&xspace); x->random();
  Vector *xtmp = new Vector(&xspace);
  Vector *b = new Vector(&bspace);
  Vector *btmp = new Vector(&bspace);
  Vector *A = new Vector(&aspace); A->zeroElement();
  
  A->random();
  float *__restrict__ A_val = ((dataInCoreFloat*)A->getDataContainer())->getFloatPtr();
  float *__restrict__ x_val = ((dataInCoreFloat*)x->getDataContainer())->getFloatPtr(); 
  float *__restrict__ xtmp_val = ((dataInCoreFloat*)xtmp->getDataContainer())->getFloatPtr(); 
  float *__restrict__ b_val = ((dataInCoreFloat*)b->getDataContainer())->getFloatPtr(); 
  float *__restrict__ btmp_val = ((dataInCoreFloat*)btmp->getDataContainer())->getFloatPtr(); 
  for(int i=0; i < 20; i++){
    A_val[i*20+i] = 5.;
  }
  Matmul *mul = new Matmul(&xspace, &bspace, A);
  mul->Forward(false,x,b);

  MKL_INT n=20, rci_request, itercount, expected_itercount=20,i;
  MKL_INT length = 128;
  double *expected_sol = new double[20];
  ConvertFtoD(x_val,expected_sol,20);
  double *rhs = new double[20];
  ConvertFtoD(b_val,rhs,20);
  double *solution = new double[20];
  double *tmp = new double[20*4];
  double *dpar= new double[128];
  MKL_INT *ipar = new MKL_INT[128];
  
  
  dcg_init(&n,solution,rhs,&rci_request,ipar,dpar,tmp);
  ipar[8] = 1;
  ipar[9] = 0;
  dpar[0] = 1.e-10;
  dcg_check(&n,solution,rhs,&rci_request,ipar,dpar,tmp);

  for(int iter = 0 ; iter < 100; iter++){
    cout<<"iter = " << iter << endl;
    dcg(&n,solution,rhs,&rci_request,ipar,dpar,tmp);
    if(rci_request == 0){
      break;
    }
    ConvertDtoF(tmp,xtmp_val,20);
    mul->Forward(false,xtmp,btmp);
    ConvertFtoD(btmp_val,&tmp[20],20);
    
  }
  dcg_get(&n,solution,rhs,&rci_request,ipar,dpar,tmp,&itercount);
  for(int iter=0;iter<20;iter++){
     cout << "  solution: "<< solution[iter] <<", expected: " << expected_sol[iter] <<endl;
  
  }
*/


  else{
  Run->SetModel(K,rho);
  Run->CreateData();
  Run->GetData(dataTrue);
  Run->SetModel(K0,rho0);
  cout << "Do line search." <<endl;
  sreed("DirectK",dK_val,sizeof(float)*nz*nx); //dK->addInv();
  sreed("modelK0",K0_val,sizeof(float)*nz*nx);
  //sreed("modelr0",rho0_val,sizeof(float)*nz*nx);
  Run->SetModel(K0,rho0);
  Run->CreateData();
  Run->GetData(dataRes);
  for(int j=0; j < nt*nx*ns; j++) dataRes[j] = dataRes[j]-dataTrue[j];
  float obj0 = ObjFuc(dataRes,nt*nx*ns);
  cout << "Initial model obj = " << obj0 << endl;  
  // First line search
  float alpha=10., objmin=1000000,alphamin;
  getch("alphamax","f",&alpha);    cout << "	alpha = " << alpha << endl;
  int mintag=0.;
  for(int i=0; i < 10; i++){
    K->linearCombo(0.,*K0,1.); //rho->linearCombo(0.,*rho0,1.);
    K->linearCombo(1.,*dK,alpha); //rho->linearCombo(1.,*drho,alpha);
    Run->SetModel(K,rho);
    Run->CreateData();
    Run->GetData(dataRes);
    for(int j=0; j < nt*nx*ns; j++) dataRes[j] = dataRes[j]-dataTrue[j];
    float obj = ObjFuc(dataRes,nt*nx*ns);
    cout << "At alpha = " << alpha << ", obj = " << obj << endl;
    if(obj < objmin){objmin = obj; alphamin=alpha;}  
    alpha = alpha*0.95;
  }
    cout << "update at alpha = " << alphamin << endl;
    K->linearCombo(0.,*K0,1.); //rho->linearCombo(0.,*rho0,1.);
    K->linearCombo(1.,*dK,alphamin); //rho->linearCombo(1.,*drho,alphamin);
  // srite("data",dataRes,sizeof(float)*ns*nx*nt);
  srite("newK",K_val,sizeof(float)*nz*nx);
  //srite("newr",rho_val,sizeof(float)*nz*nx);
  }


  // Clean up

   std::cout << "========== Done, Cleaning up ==========" <<std::endl; 

  return 0;
}