#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"
#include <omp.h>
#include "base/utilities.h"

#include "base/Vector.h"
#include "base/inCoreFloat.h"
#include "base/DataInCoreFloat.h"
#include "base/Operator.h"
#include "base/MultiOperator.h"

#include "CreateModel.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Prop.h"

using namespace std;
using namespace SEP;

/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/


int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "         Full Waveform inversion.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(1);   // Initialize the parallel environment
  


  int nz=201, nx=501;
  int choice;
  getch("type","d",&choice);   
  float *__restrict__ K = new float[nz*nx];   
  float *__restrict__ rho = new float[nz*nx];
  float modulus = 11.52;
  float density = 2.;
  
  // Background
  for(int i1=0; i1<nz;i1++){
    for(int i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K[i] = modulus;
      rho[i] = density;
    }
  }   
  
  int z0,z1,z2,z3,z4,z5;
  z0=0; z1=80; z2=110; z3=130; z4=148;
  // Add layer
  for(llong i1=z1; i1<z2;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K[i] = modulus*1.1;
    }
  } 
  for(llong i1=z3; i1<z4;i1++){
    for(llong i2=0; i2<nx; i2++){
      llong i = i1*nx+i2;
      K[i] = modulus*1.1;
    }
  }  
  srite("mod0",K,sizeof(float)*nx*nz);
  
  if(choice == 0|| choice==2){
  cout << "	=====Layer model=====" << endl; 
  int x0=190,x1=310;   
  for(llong i1=z1; i1<z1+5;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K[i] = K[i]*1.03;
    }
  } 
  for(llong i1=z1+5; i1<z1+10;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K[i] = K[i]*0.97;
    }
  } 
  for(llong i1=z1+10; i1<z1+15;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K[i] = K[i]*1.03;
    }
  } 
  for(llong i1=z1+15; i1<z1+20;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K[i] = K[i]*0.97;
    }
  } 
  for(llong i1=z1+20; i1<z1+25;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K[i] = K[i]*1.03;
    }
  } 
  for(llong i1=z1+25; i1<z1+30;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K[i] = K[i]*0.97;
    }
  }   
  
  }
  
  if(choice == 1){
  cout << "	=====Pinch model=====" << endl;
  // Pinch 1
  int x0=230,x1=290;   
  int xstart=190, xend=310;
  int ztop = z1, zbot = z1+6;
  float slope = 6./(x1-x0);
  float posz;
  for(llong i1=ztop; i1<zbot;i1++){
    for(llong i2=xstart; i2<xend; i2++){
      posz = slope*(i2-x0) + ztop;
      if( (float)i1 < posz){
      llong i = i1*nx+i2;
      K[i] = K[i]*1.03;
      }
      else{
      llong i = i1*nx+i2;
      K[i] = K[i]*0.97;      
      }
    }
  }  

  // Pinch 2
  x0=210,x1=300; 
  ztop = z1+6;
  zbot = z1+12;  
  slope = -6./(x1-x0);
  for(llong i1=ztop; i1<zbot;i1++){
    for(llong i2=xstart; i2<xend; i2++){
      posz = slope*(i2-x1) + ztop;
      if( (float)i1 < posz){
      llong i = i1*nx+i2;
      K[i] = K[i]*1.02;
      }
      else{
      llong i = i1*nx+i2;
      K[i] = K[i]*0.98;      
      }
    }
  }  
 
  // Pinch 3
  x0=190,x1=310; 
  ztop = z1+12;
  zbot = z1+18;  
  slope = 6./(x1-x0);
  for(llong i1=ztop; i1<zbot;i1++){
    for(llong i2=xstart; i2<xend; i2++){
      posz = slope*(i2-x0) + ztop;
      if( (float)i1 < posz){
      llong i = i1*nx+i2;
      K[i] = K[i]*1.03;
      }
      else{
      llong i = i1*nx+i2;
      K[i] = K[i]*0.97;      
      }
    }
  } 
  // Pinch 4
  x0=200,x1=250; 
  ztop = z1+18;
  zbot = z1+24;  
  slope = -6./(x1-x0);
  for(llong i1=ztop; i1<zbot;i1++){
    for(llong i2=xstart; i2<xend; i2++){
      posz = slope*(i2-x1) + ztop;
      if( (float)i1 < posz){
      llong i = i1*nx+i2;
      K[i] = K[i]*1.02;
      }
      else{
      llong i = i1*nx+i2;
      K[i] = K[i]*0.98;      
      }
    }
  } 
  // Pinch 5
  x0=250,x1=300; 
  ztop = z1+24;
  zbot = z1+30;  
  slope = 6./(x1-x0);
  for(llong i1=ztop; i1<zbot;i1++){
    for(llong i2=xstart; i2<xend; i2++){
      posz = slope*(i2-x0) + ztop;
      if( (float)i1 < posz){
      llong i = i1*nx+i2;
      K[i] = K[i]*1.03;
      }
      else{
      llong i = i1*nx+i2;
      K[i] = K[i]*0.97;      
      }
    }
  } 
  } // end of if
  
  
  if(choice==2){
  cout << "	=====Pipe model=====" << endl;
  int x0=236,x1=244;   
  for(llong i1=z1; i1<z2;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K[i] = modulus*1.12;
    }
  } 
  x0=272,x1=275; 
  for(llong i1=z1; i1<z2;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K[i] = modulus*1.12;
    }
  }     
  }
  
  srite("mod",K,sizeof(float)*nx*nz);
  /*
 
  llong x0=120,x1=180;
  for(llong i1=z1; i1<z1+5;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*1.03;
    }
  } 
  for(llong i1=z1+5; i1<z1+10;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*0.97;
    }
  } 
  for(llong i1=z1+10; i1<z1+15;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*1.03;
    }
  } 
  for(llong i1=z1+15; i1<z1+20;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*0.97;
    }
  } 
  for(llong i1=z1+20; i1<z1+25;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*1.03;
    }
  } 
  for(llong i1=z1+25; i1<z1+30;i1++){
    for(llong i2=x0; i2<x1; i2++){
      llong i = i1*nx+i2;
      K_val[i] = K_val[i]*0.97;
    }
  } 
}
*/


    return 0;
}