#ifndef PROP_H
#define PROP_H 1
#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "base/Vector.h"

#include "Taper.h"
#include "Clean.h"
#include <cmath>
#include <iostream>


#include "TimeMarch.h"

typedef long long llong;

/*
	Created by Yinbin Ma  
	Wave propagation module.
	Designed for single shot propagation.
	Can handle point source and Born source
	Use carefully and have fun.
	
	First created: 2015-01-01
	Last Modified: 2015-07-06
*/	

namespace SEP {

class Prop: public Operator{
  public:
    Prop(Space *mod, Space *data){
      setDomain(mod);
      setRange(data);
      Utemp = new Vector((inCoreFloat *)data);
      init();
      
    }
    ~Prop();
    void SetPropPar(llong _nb, float dt, float dx,Vector *K);
    void SetModel(Vector *K, Vector *rho);
    virtual bool Forward(bool add, Vector *model, Vector *data);    // Forward propagation
    virtual bool Adjoint(bool add, Vector *model, Vector *data);    // Adjoint propagation
void Dot(){
      Vector m1(domain);//=new Vector(domain);
      Vector d1(range);//=new Vector(range);
      Vector m2(domain);//=new Vector(domain);
      Vector d2(range);//=new Vector(range);
      float *__restrict__ m1val = ((dataInCoreFloat*)m1.getDataContainer())->getFloatPtr();
      float *__restrict__ d1val = ((dataInCoreFloat*)d1.getDataContainer())->getFloatPtr();
      float *__restrict__ m2val = ((dataInCoreFloat*)m2.getDataContainer())->getFloatPtr();
      float *__restrict__ d2val = ((dataInCoreFloat*)d2.getDataContainer())->getFloatPtr();
      m1.random();
      d2.random();

      Forward(false,&m1,&d1);
  
      double dot1=0; //d1.innerProd(d2);
      for(llong i=mblock; i<(nt-1)*mblock;i++)dot1+=d1val[i]*d2val[i];

      Adjoint(false,&m2,&d2);
      double dot2=0; //m2.innerProd(m1);
      for(llong i=mblock; i<(nt-1)*mblock;i++)dot2+=m1val[i]*m2val[i];
      std::cout<<"Dot product test with add=false: \n"<<dot1<<"\n"<<dot2<<"\n";

      Forward(true,&m1,&d1);
      double dot3=0; //d1.innerProd(d2);
      for(llong i=mblock; i<(nt-1)*mblock;i++)dot3+=d1val[i]*d2val[i];
      Adjoint(true,&m2,&d2);
      double dot4=0.; //m2.innerProd(m1);
      for(llong i=mblock; i<(nt-1)*mblock;i++)dot4+=m1val[i]*m2val[i];


     std::cout<<"Dot product test with add=true: \n"<<dot3<<"\n"<<dot4<<"\n";

    }
  private:
    void init();
    llong nz,nx,nt, mblock, nb;
    Vector *pre, *cur, *next, *temp;
    Vector *KTX, *rho;                        // Why do I need rho here?
    TimeMarch *Tmarch;
    float dt2;
    inCoreFloat *mspace;
    void PullData(bool add, Vector *mvec, Vector *uvec, llong position);
    void PushData(bool add, Vector *mvec, Vector *uvec, llong position);
    void CleanBound(Vector *mvec, Vector *uvec);               // use for high precision dot test
    
    CleanVec Zero;
    Taper *taper;
    Vector *Utemp;

};

void Prop::SetPropPar(llong _nb, float dt, float dx,Vector *K){
  nb = _nb;
  Tmarch = new TimeMarch(nb, dt, dx, K);    // Set up time marching operator
  mspace = (inCoreFloat *)K->getSpace();    // get space for the model
  KTX = new Vector(K);                      // KTX is used for scaling the source
  pre    = new Vector(K);                   // temp vector
  cur    = new Vector(K);                   // temp vector
  next   = new Vector(K);                   // temp vector
  taper = new Taper(K, 55, 55, 0.97); // taper operator
  dt2 = 1.*dt*dt;                           // dt^2 notation
}

 void Prop::SetModel(Vector *K, Vector *rho){
   Tmarch->SetModel(K,rho);                  // Change model 
   KTX->linearCombo(0.,*K,dt2);            // change K here
 }

void Prop::init() {
  inCoreFloat *in_space = (inCoreFloat *)domain;
  std::vector<long long> in_dims = in_space->get_ns();
  nt = in_dims[0];
  nz = in_dims[1];
  nx = in_dims[2];
  mblock = nx*nz;
}

Prop::~Prop(){
  delete pre;
  delete cur;
  delete next;
}

bool Prop::Forward(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(data,nt,mblock);
  llong cur_pos = 0;      // shift in the data cube
  Zero.Zero(pre,nz,nx);
  Zero.Zero(cur,nz,nx);
  Zero.Zero(next,nz,nx);
  Zero.Zero(Utemp, nt, mblock);
  CleanBound(model, Utemp);
  //PullData(true,pre,model,cur_pos);
  //PullData(true,cur,model,cur_pos);
  //PushData(add,cur,data,cur_pos);
  PushData(add,next,data,cur_pos);  // write u(0)
  cur_pos = cur_pos + nz*nx;
  PushData(add,next,data,cur_pos);  // write u(0) 
  for(llong it = 1; it < nt-1; it++){
    Tmarch->FMarch(pre,cur,next);         // Time marching
    PullData(true,next,Utemp,cur_pos);
    taper->tapeISPC(cur,next);              // We need to taper both!
    cur_pos = cur_pos + nz*nx;
    PushData(add,next,data,cur_pos);
    temp = pre; pre = cur; cur = next; next = temp;
  }
  
  return 0;
}

bool Prop::Adjoint(bool add, Vector *model, Vector *data){

  if(!add) Zero.Zero(model,nt,mblock);
  llong cur_pos = (nt-1)*nz*nx;
  Zero.Zero(pre,nz,nx);
  Zero.Zero(cur,nz,nx);
  Zero.Zero(next,nz,nx);
  PushData(false,next,Utemp,cur_pos);  // write u(nt-1)
  cur_pos = cur_pos - nz*nx;               // move to next time slice
  PushData(false,next,Utemp,cur_pos);  // write u(nt-1)
  //PullData(true,pre,data,cur_pos);
  //PullData(true,cur,data,cur_pos);
  //PushData(add,cur,model,cur_pos);

    
  for(llong it = 1; it < nt-1; it++) {
    // time marching for it+1 solution
    Tmarch->BMarch(pre,cur,next);         // Time marching
    PullData(true,next,data,cur_pos);  //directly add source 
    taper->tapeISPC(cur,next);              // We need to taper both!
    cur_pos = cur_pos - nz*nx;               // move to next time slice
    PushData(false,next,Utemp,cur_pos);         // save u(it+1)
    temp = pre; pre = cur; cur = next; next = temp;
  }
  CleanBound(Utemp,model);
  return 0;
}

// The following is ugly
void Prop::PullData(bool add, Vector * mvec, Vector * uvec, llong position) {
  float *__restrict__ mval = ((dataInCoreFloat*)mvec->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)uvec->getDataContainer())->getFloatPtr();
  float *__restrict__ Ks = ((dataInCoreFloat*)KTX->getDataContainer())->getFloatPtr();

  if(add){
    #pragma omp parallel for
    for(long long i = 0; i < nz*nx; i++) {
      mval[i] += uval[i + position]*Ks[i];
    }
  }
  else { 
    #pragma omp parallel for
    for(long long i = 0; i < nz*nx; i++) {
      mval[i] = uval[i + position]*Ks[i];
    }
  }
}

void Prop::PushData(bool add, Vector * mvec, Vector * uvec, llong position) {
  float *__restrict__ mval = ((dataInCoreFloat*)mvec->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)uvec->getDataContainer())->getFloatPtr();
  if(add){
  
   #pragma omp parallel for
    for(long long i = 0; i < nz*nx; i++) {
      uval[i + position] += mval[i];
    }
  }
  else {
    #pragma omp parallel for
    for(long long i = 0; i < nz*nx; i++) {
      uval[i + position] = mval[i];
    }
  }
}

void Prop::CleanBound(Vector *mvec, Vector *uvec){
  float *__restrict__ mval = ((dataInCoreFloat*)mvec->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)uvec->getDataContainer())->getFloatPtr();
  #pragma omp parallel for
  for(llong it=0; it < nt; it++){
    for(long long iz = nb; iz < nz-nb-1; iz++) {
      for(llong ix = nb; ix < nx - nb-1; ix++) {
        llong position = it*nx*nz + iz*nx+ix;
        uval[position] += mval[position];
      }
    }
  }
}


}

#endif