#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"
#include <omp.h>
#include "base/utilities.h"

#include "base/Vector.h"
#include "base/inCoreFloat.h"
#include "base/DataInCoreFloat.h"
#include "base/Operator.h"
#include "base/MultiOperator.h"

#include "CreateModel.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Prop.h"

using namespace std;
using namespace SEP;

/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/



int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "         Full Waveform inversion.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(8);   // Initialize the parallel environment
  
  // Define dimension of interest
  long long nx, nz, nt;
  long long sx, sz,ns,ox,dxs;
  long long boundary;
  float dx, dz, dt;
  float velocity, density, modulus;
  float fmax=15;
  nt = 1500;
  nx = 401;
  nz = 201;
  ns = 1;
  sx = 200;
  sz = 65;
  boundary = 10;

  density=2.;          // [g/cm^3]
  velocity=2.4;        // [km/s]
  dx = 10.; dz = 10.;  // [m]
  dt = 1.5;            // [ms]
  modulus=velocity*velocity*density; // [GPa]

  // Set up space 
  inCoreFloat1D sspace(nt);               // space for shot
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wave field cube
  inCoreFloat3D dspace(ns,nt,nx);         // space for data space
  // Set up vectors
  Vector *K      = new Vector(&mspace);    K->zeroElement();      // migration velocity
  Vector *rho    = new Vector(&mspace);    rho->zeroElement();     // Velocity vector
  Vector *GradK  = new Vector(&mspace);    GradK->zeroElement();      // migration velocity
  Vector *Gradr  = new Vector(&mspace);    Gradr->zeroElement();     // Velocity vector
  Vector *vel    = new Vector(&mspace);    vel->zeroElement();      // migration velocity
  Vector *shot   = new Vector(&sspace);    shot->zeroElement();     // shot profile
  Vector *data   = new Vector(&dspace);    data->zeroElement();     // data
  Vector *Syndata= new Vector(&dspace);    Syndata->zeroElement();     // data
  Vector *u1     = new Vector(&uspace);    u1->zeroElement();       // Wave field
  Vector *u2     = new Vector(&uspace);    u2->zeroElement();       // Wave field  
  Vector *u3     = new Vector(&uspace);    u3->zeroElement();       // Wave field  


  // Read in source wavelet and velocity field
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  sreed("wavelet",shot_val,sizeof(float)*nt); 
  
  // Set up model
  Models *CModel =  new Models(nz, nx, modulus, density);
  CModel->Model0(K,rho);


  // Check numerical condition
  cout << "    Dispersion: Vmin/(fmax*dx) = " << 1000.*velocity/fmax/dx <<", should be > 4"<<endl;
  cout << "    Courant:    Vmax*dt/dx     = " << velocity*dt/dx <<", should be < 0.717"<<endl;
 
  // Generate operators for wave progation
  Prop *Lmig = new Prop(&uspace,&uspace);
  Lmig->SetPropPar(boundary,dt, dx, K);
  Lmig->SetModel(K,rho);

  Prop *Lmig0 = new Prop(&uspace,&uspace);
  Lmig0->SetPropPar(boundary,dt, dx, K);
  CModel->ModelC(K,rho);
  Lmig0->SetModel(K,rho);
  //Lmig->DotTest(true);
  
  // Inject source wavelet
  Source *Ss = new Source(nt, nz, nx);
  Ss->Injection(shot, u1, sz, sx);
  
  // Forward propagation
 // double t0 = timer();
  //Lmig0->Forward(false,u1,u3);         // source -> ufield
  //u3->addInv();
  Lmig->Forward(true,u1,u3);         // source -> ufield
//  double t1 = timer(); 
 // cout << "   wave propagation, takes " << t1-t0 << "seconds. " <<endl;
  
  // Save data
  SensorPad *SP = new SensorPad(&dspace, &uspace, sz, boundary);
  SP->SetShot(0);  
  SP->Adjoint(false,data,u3);
  
  // Output data  
  float *__restrict__ u3_val = ((dataInCoreFloat*)u3->getDataContainer())->getFloatPtr();
  float *__restrict__ d_val = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr(); 
  srite("wave1",u3_val,sizeof(float)*nz*nx*nt);
  srite("data1",d_val,sizeof(float)*nx*nt); 
  return 0;
  // now let us do some simulation
  float dVp = 1./sqrt(1.2)-1.;
  float drho = 0.2;
  float R0 = 0.5*(dVp+drho);
  float Rt;
  float G = 0.5*dVp;
  float t0 = 100;
  float depth = (110-sz);
  float length;
  float theta;
  float *__restrict__ Synd_val = ((dataInCoreFloat*)Syndata->getDataContainer())->getFloatPtr();
  for(llong ih=0; ih < 200; ih++){
    length = sqrt(ih*ih+depth*depth);
    theta = asin(ih/length);
    Rt = R0 + G*pow(sin(theta),2)*(1.+pow(tan(theta),2));
    for(llong it=0; it < nt; it++) {
      float universal = (it-t0)*dt-length*10/velocity;
      Synd_val[it*nx+ih+200] = sqrt(cos(theta))*Rt*exp(-1.*universal*universal/1000.)*universal;
      Synd_val[it*nx-ih+200] = Synd_val[it*nx+ih+200];
    }
  }
  srite("Syndata1",Synd_val,sizeof(float)*nx*nt); 
  //return 0;

  // Second model
  CModel->Model1(K,rho);
  Lmig->SetModel(K,rho);
  Lmig0->Forward(false,u1,u3);         // source -> ufield
  u3->addInv();
  Lmig->Forward(true,u1,u3);         // source -> ufield
  SP->Adjoint(false,data,u3);
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr(); 
  srite("model",K_val,sizeof(float)*nx*nz); 
  srite("wave2",u3_val,sizeof(float)*nz*nx*nt);
  srite("data2",d_val,sizeof(float)*nx*nt);    
  // Clean up

  dVp = 1./sqrt(1.2)-1.;
  drho = 0.0;
  R0 = 0.5*(dVp+drho);
  G = 0.5*dVp;
  t0 = 100;
  depth = (110-sz);

    for(llong ih=0; ih < 200; ih++){
    length = sqrt(ih*ih+depth*depth);
    theta = asin(ih/length);
    Rt = R0 + G*pow(sin(theta),2)*(1.+pow(tan(theta),2));
    for(llong it=0; it < nt; it++) {
      float universal = (it-t0)*dt-length*10/velocity;

      Synd_val[it*nx+ih+200] = sqrt(cos(theta))*Rt*exp(-1.*universal*universal/1000.)*universal;
      Synd_val[it*nx-ih+200] = Synd_val[it*nx+ih+200];
    }
  }
  srite("Syndata2",Synd_val,sizeof(float)*nx*nt); 
  
  
  
   std::cout << "========== Done, Cleaning up ==========" <<std::endl; 

  return 0;
}