#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"

#include <omp.h>
#include "base/utilities.h"

#include "base/s_reg.h"
#include "base/hypercube.h"
#include "base/Vector.h"
#include "base/inCoreFloat.h"
#include "base/DataInCoreFloat.h"
#include "base/Operator.h"
#include "base/MultiOperator.h"

#include "Interpolation.h"
#include "CreateModel.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Prop.h"
#include "Imaging.h"
#include "Engine.h"
extern "C" {
#include "sep3d.h"
}
using namespace std;
using namespace SEP;

/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/
extern "C"  int sep_get_data_axis_par(const char *r,int *ax,int *ff,float *o,float *d,char *l);

float ObjFuc(float *dataRes, int len){
  float obj=0.;
  for(int i=0; i < len; i++){
    obj += dataRes[i]*dataRes[i];
  }
  return obj;
}



int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "         Full Waveform inversion.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(10);   // Initialize the parallel environment
  
  // Define dimension of interest
  int nx, nz, nt;
  int init;
  int sx, sz,ns,oxs,dxs;
  int boundary;
  float dx, dz, dt;
  float velocity, density, modulus;
  float fmax=15;
  int nb, bsize;

  // Get parameters
  getch("nb","d",&nb);    cout << "	nb = " << nb << endl;
  getch("bsize","d",&bsize);    cout << "	bsize = " << bsize << endl;
  getch("nx","d",&nx);    cout << "	nx = " << nx << endl;
  getch("nz","d",&nz);    cout << "	nz = " << nz << endl;
  getch("nt","d",&nt);    cout << "	nt = " << nt << endl;
  getch("boundary","d",&boundary);    cout << "	boundary = " << boundary << endl;
  getch("dx","f",&dx);    cout << "	dx = " << dx <<"[m]"<< endl;
  dz = dx;
  getch("sz","d",&sz);    cout << "	sz = " << sz << endl;
  getch("oxs","d",&oxs);  cout << "	oxs = " << oxs << endl;
  getch("dxs","d",&dxs);  cout << "	dxs = " << dxs << endl;
  getch("dt","f",&dt);    cout << "	dt = " << dt << "[ms]"<< endl;
  getch("ns","d",&ns);    cout << "	ns = " << ns << endl;
  getch("density","f",&density); cout << "	density = " << density << "[g/cm^3]" <<endl;
  getch("velocity","f",&velocity); cout << "	velocity = " << velocity << "[km/s]" <<endl;
  getch("init","d",&init);    cout << "	init = " << init << endl;
  modulus=velocity*velocity*density;  cout << "	modulus = " << modulus << "[GPa]" <<endl; 

  // Set up space 
  inCoreFloat1D sspace(nt);               // space for shot
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wave field cube
  inCoreFloat2D dspace(nt,nx);            // space for data space

  // Set up vectors
  Vector *K      = new Vector(&mspace);    K->zeroElement();      // migration velocity
  Vector *rho    = new Vector(&mspace);    rho->zeroElement();     // Velocity vector
  Vector *dK      = new Vector(&mspace);    dK->zeroElement();      // migration velocity
  Vector *drho    = new Vector(&mspace);    drho->zeroElement();     // Velocity vector
  Vector *K0      = new Vector(&mspace);    K0->zeroElement();      // migration velocity
  Vector *rho0    = new Vector(&mspace);    rho0->zeroElement();     // Velocity vector
  Vector *HK      = new Vector(&mspace);    HK->zeroElement();      // migration velocity
  Vector *Hrho    = new Vector(&mspace);    Hrho->zeroElement();     // Velocity vector
  Vector *shot   = new Vector(&sspace);    shot->zeroElement();     // shot profile

  // Read in wavelet and construct model
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  sreed("wavelet",shot_val,sizeof(float)*nt); 
  // Set up source 
  std::vector<int> sxv;
  std::vector<int> szv;
  for(int is=0; is < ns; is++){
    sxv.push_back(oxs+is*dxs);
  //  sxv.push_back(150);
    szv.push_back(sz);
  }

  // Set up model
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr(); 
  float *__restrict__ K0_val = ((dataInCoreFloat*)K0->getDataContainer())->getFloatPtr();
  float *__restrict__ rho0_val = ((dataInCoreFloat*)rho0->getDataContainer())->getFloatPtr();
  float *__restrict__ dK_val = ((dataInCoreFloat*)dK->getDataContainer())->getFloatPtr();
  float *__restrict__ drho_val = ((dataInCoreFloat*)drho->getDataContainer())->getFloatPtr(); 

   
  // Set up Engine
  Engine *Run = new Engine(&mspace,&dspace,&uspace,shot,sxv,szv);
  Run->SetPropPar(nb, bsize, boundary, dt, dx, dz); 
  float *dataTrue=new float[ns*nx*nt];
  float *wave=new float[nz*nx*nt];
  float *dataDirect=new float[ns*nx*nt]; 
  float *dataRes=new float[ns*nx*nt]; 
  float *dataRes1=new float[ns*nx*nt];
  float *dataRes2=new float[ns*nx*nt];
  float *dataRes3=new float[ns*nx*nt];
  if( init==0){
    Models *CModel =  new Models(nz, nx, modulus, density);
    CModel->Model9(K,rho);
    CModel->Model8(K0,rho0); 
    sreed("TrueV", K_val,sizeof(float)*nz*nx);
    sreed("MigV", K0_val,sizeof(float)*nz*nx); 
    for(llong i=0; i < nx*nz;i++){
      K_val[i] = K_val[i]*K_val[i]*rho_val[i]/1000000.;
      K0_val[i] = K0_val[i]*K0_val[i]*rho0_val[i]/1000000.;
    }   
    srite("modelK",K_val,sizeof(float)*nz*nx);
    srite("modelr",rho_val,sizeof(float)*nz*nx); 
    srite("modelK0",K0_val,sizeof(float)*nz*nx);
    srite("modelr0",rho0_val,sizeof(float)*nz*nx);
    //CModel->ModelC(K,rho);
    Run->SetModel(K,rho);
    Run->CreateData();
    Run->GetWave(wave, 5);
    Run->GetData(dataDirect);
    srite("dataDirect",dataDirect,sizeof(float)*ns*nt*nx);
    //for(int i=0; i < nt*nx*ns; i++) dataTrue[i] = dataTrue[i]-dataDirect[i];
    srite("wave",wave,sizeof(float)*nz*nt*nx);
    return 0;    
  }
  else{
    sreed("modelK",K_val,sizeof(float)*nz*nx);
    sreed("modelr",rho_val,sizeof(float)*nz*nx); 
    sreed("modelK0",K0_val,sizeof(float)*nz*nx);
    sreed("modelr0",rho0_val,sizeof(float)*nz*nx);
    sreed("dataTrue",dataTrue,sizeof(float)*ns*nt*nx); 
    srite("dataDirect",dataDirect,sizeof(float)*ns*nt*nx);   
  }


  // Collect data residual and inject data residual
  cout << "start to run " <<endl;
  Run->SetModel(K,rho);
  Run->CreateData();
  Run->GetData(dataTrue);
  Run->SetModel(K0,rho0);
  Run->CreateData();
  Run->GetData(dataRes);
  for(int i=0; i < nt*nx*ns; i++) dataRes[i] = dataRes[i]-dataTrue[i];
  float obj0 = ObjFuc(dataRes,nt*nx*ns);
  cout << "Initial model obj = " << obj0 << endl;   
  Run->PutData(dataRes);
  
  // Iterative steepest descent
  float r0, r1, r2, r3,rmin;
  int Nfunc=0,niter=1,success=0;
  float alpha=4, objmin=1000000,alphamin;
    getch("alphamax","f",&alpha);    cout << "	alpha = " << alpha << endl;
  float *objs=new float[niter];
  float *alphas=new float[niter];
  float *Klong = new float[nx*nz*niter];
  
  
  for( int iter=0; iter < niter; iter++){
    cout << " Iteration :" << iter <<endl;
    K->linearCombo(0.,*K0,1.); //rho->linearCombo(0.,*rho0,1.);
    //K->linearCombo(1.,*dK,0.5*alpha); //rho->linearCombo(1.,*drho,alpha);
    Run->SetModel(K,rho);
    Run->CreateData();
    Run->GetData(dataRes);
    for(int j=0; j < nt*nx*ns; j++) dataRes[j] = dataRes[j]-dataTrue[j];    
    r0 = ObjFuc(dataRes,nt*nx*ns);
    Nfunc = Nfunc+1;
    Run->PutData(dataRes);
    Run->GradientFunc(dK,drho,K0,rho0);
    
    // Line Search
    for(int i=0; i < 8; i++){
      K->linearCombo(0.,*K0,1.); //rho->linearCombo(0.,*rho0,1.);
      K->linearCombo(1.,*dK,alpha); //rho->linearCombo(1.,*drho,alpha);
      Run->SetModel(K,rho);
      Run->CreateData();
      Run->GetData(dataRes);
      for(int j=0; j < nt*nx*ns; j++) dataRes[j] = dataRes[j]-dataTrue[j];
      float obj = ObjFuc(dataRes,nt*nx*ns);
      cout << "      At alpha = " << alpha << ", obj = " << obj << endl;
      if(obj < objmin){objmin = obj; alphamin=alpha;}  
      alpha = alpha*0.8;
    }
    cout << "  Obj0 = " << r0 << ", ObjNew = " << objmin << " , alpha = " << alphamin <<", Nfunc = " << Nfunc<< endl;
    K0->linearCombo(1.,*dK,alphamin); 
    objs[iter] = objmin; alphas[iter] = alphamin;
    alpha=alphamin*1.5;
    objmin=1000000.;
}

/*
  for( int iter=0; iter < niter; iter++){
    cout << " Iteration :" << iter <<endl;
    if(alpha <0.0002) alpha = 0.2;
    r0 = ObjFuc(dataRes,nt*nx*ns);
    Run->PutData(dataRes);
    Run->GradientFunc(dK,drho,K0,rho0);
    Nfunc = Nfunc+2;
    // Line Search
    success=0;
    for( int i=0; i < 5; i++){
      // phi(0.5 alpha)
      K->linearCombo(0.,*K0,1.); //rho->linearCombo(0.,*rho0,1.);
      K->linearCombo(1.,*dK,0.5*alpha); //rho->linearCombo(1.,*drho,alpha);
      Run->SetModel(K,rho);
      Run->CreateData();
      Run->GetData(dataRes1);
      for(int j=0; j < nt*nx*ns; j++) dataRes1[j] = dataRes1[j]-dataTrue[j];    
      r1 = ObjFuc(dataRes1,nt*nx*ns);
      Nfunc = Nfunc+1;
      // phi(alpha)
      K->linearCombo(0.,*K0,1.); //rho->linearCombo(0.,*rho0,1.);
      K->linearCombo(1.,*dK,alpha); //rho->linearCombo(1.,*drho,alpha);
      Run->SetModel(K,rho);
      Run->CreateData();
      Run->GetData(dataRes2);
      for(int j=0; j < nt*nx*ns; j++) dataRes2[j] = dataRes2[j]-dataTrue[j];    
      r2 = ObjFuc(dataRes2,nt*nx*ns);
      Nfunc = Nfunc+1;
      // phi(1.5 alpha)
      K->linearCombo(0.,*K0,1.); //rho->linearCombo(0.,*rho0,1.);
      K->linearCombo(1.,*dK,1.5*alpha); //rho->linearCombo(1.,*drho,alpha);
      Run->SetModel(K,rho);
      Run->CreateData();
      Run->GetData(dataRes3);
      for(int j=0; j < nt*nx*ns; j++) dataRes3[j] = dataRes3[j]-dataTrue[j];    
      r3 = ObjFuc(dataRes3,nt*nx*ns);
      Nfunc = Nfunc+1;
      cout << "     r1=" << r1 << ", r2=" << r2 << ", r3=" << r3 << endl;
      // judge
      if(r0 < r1 && r0 < r2 && r0 < r3){ 
        alpha = 0.25*alpha;
      }
      else if(r3 < r1 && r3 < r2){
        alpha = 1.5*alpha;
        rmin=r3;
        success=1;
        memcpy(dataRes, dataRes3, sizeof(float)*ns*nx*nt);
        break;
      }
      else if(r1 < r2 && r1 < r3){
        alpha = 0.5*alpha;
        rmin=r1;
        success=1;
        memcpy(dataRes, dataRes1, sizeof(float)*ns*nx*nt);
        break;
      }
      else if(r2 < r1 && r2 < r3){
        rmin=r2;
        success=1;
        memcpy(dataRes, dataRes2, sizeof(float)*ns*nx*nt);
        break;
      }

    }

    // update and print to screen
    if(success==1){
      cout << "  Obj0 = " << r0 << ", ObjNew = " << rmin << " , alpha = " << alpha <<", Nfunc = " << Nfunc<< endl;
      r0=rmin;  
      K0->linearCombo(1.,*dK,alpha); //rho->linearCombo(0.,*rho0,1.);
      objs[iter] = r0; alphas[iter] = Nfunc;
    }
    else{
      cout << "  Failed." << endl;
      cout << "  Obj0 = " << r0 << ", ObjNew = " << r2 << " , alpha = " << alpha <<", Nfunc = " << Nfunc<< endl;
      alpha=alpha*4.; r0=r2;
      objs[iter] = r0; alphas[iter] = alpha;
      K0->linearCombo(1.,*dK,alpha);      
      memcpy(dataRes, dataRes2, sizeof(float)*ns*nx*nt);
    }
    llong shift=nx*nz*iter;
    for(llong i=0; i < nx*nz; i++){
      Klong[shift+i] = K0_val[i];
    }
   // alpha = 2*alpha;
  }
*/

  srite("GradK",dK_val,sizeof(float)*nz*nx);

  float *__restrict__ HK_val = ((dataInCoreFloat*)HK->getDataContainer())->getFloatPtr();
  float *__restrict__ Hrho_val = ((dataInCoreFloat*)Hrho->getDataContainer())->getFloatPtr();   
  cout << "norm dK = " << dK->innerProd(*dK) << endl; 
  srite("data",dataRes,sizeof(float)*ns*nx*nt);
  srite("newK",K0_val,sizeof(float)*nz*nx);
  srite("newr",rho_val,sizeof(float)*nz*nx);
  srite("Klong",Klong,sizeof(float)*nz*nx*niter);
  srite("objs",objs,sizeof(float)*niter);
  srite("alphas",alphas,sizeof(float)*niter);
  // Clean up

   std::cout << "========== Done, Cleaning up ==========" <<std::endl; 

  return 0;
}