#ifndef SENSORPAD_H
#define SENSORPAD_H 1
#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "Clean.h"
typedef long long llong;
/*
Sensor padding operator
  Use to injector data[nt][nx] to source[nt][nz][nx] 
  With fixed depth sz.
  Need to set shot number each time calling this function.
  There is no confict of memory address in the for loop. 
  Use carefully and have fun.

Created:       2015-01-15
Last modified: 2015-08-03

*/

namespace SEP {

class SensorPad: public Operator{
  
  public:
    SensorPad(Space *mod, Space *data,llong _sz,llong _bnx=0){
      setDomain(mod); 
      setRange(data);
      init(_sz, _bnx);
    };
    void SetShot(llong _sx) {sx = _sx;}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);

  private:
    void init(llong sz, llong bnx);
    llong nt,nz,nx;
    llong sz, bnx,sx;
    CleanVec Zero;
};





void SensorPad::init(llong _sz,llong _bnx) {
   inCoreFloat *wave = (inCoreFloat *)range;
   std::vector<long long> in_dims = wave->get_ns();
   nt = in_dims[0];
   nz = in_dims[1];
   nx = in_dims[2];
   sz = _sz;
   bnx = _bnx;
   sx = 200;
   std::cout << "nt = " << nt << ", nx = " << nx << ", nz = "<< nz <<std::endl;
}


bool SensorPad::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,nt,nz*nx);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
 
   llong block = nx*nz;
  llong it,ix;
  llong nm=0;
  //#pragma omp parallel for private(it,ix)
  for(it=1; it<nt-1; it++){
    for(ix=bnx; ix<sx-nm; ix++){
      d[it*block + sz*nx+ix] += m[it*nx + ix];
    }
    for(ix=sx+nm; ix<nx-bnx; ix++){
      d[it*block + sz*nx+ix] += m[it*nx + ix];
    }
  }
  return 0;

}

bool SensorPad::Adjoint(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(model,nt,nx);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  llong block = nx*nz; 
  llong it,ix;
  llong nm=0;
  //#pragma omp parallel for private(it,ix)
  for(it=1; it<nt-1; it++){
    for(ix=bnx; ix<sx-nm; ix++){
      m[it*nx + ix] += d[it*block + sz*nx+ix];
    }
    for(ix=sx+nm; ix<nx-bnx; ix++){
      m[it*nx + ix] += d[it*block + sz*nx+ix];
    }
  }
  return 0;
} 


}
#endif
