#ifndef SOURCE_H
#define SOURCE_H 1
#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "base/Vector.h"
#include <cmath>
#include <omp.h>
#include <iostream>
typedef long long llong;
/*
Model creation module
Used to simplify the main code.
It is not an operator yet.

Created:       2015-07-06
Last modified: 2015-08-03

*/

namespace SEP {

class Source{
  
  public:
    Source(llong _nt, llong _nz, llong _nx){
      nt = _nt;
      nz = _nz;
      nx = _nx;
    }
    ~Source(){}
    void Injection(Vector *s, Vector *u, llong sz, llong sx);
    void InjectionDipole(Vector *s, Vector *u, llong sz, llong sx);
    void BInjection(Vector *s, Vector *u, llong sz, llong sx);
    void InjectionLine(Vector *s, Vector *u, llong sz, llong sx);
  private:
    llong nt, nz,nx;
};



void Source::Injection(Vector *shot, Vector *u, llong sz, llong sx) {
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr();   
  float *__restrict__ u_val = ((dataInCoreFloat*)u->getDataContainer())->getFloatPtr();  

  for(llong i=1; i<nt-1;i++){
    u_val[i*nx*nz+sz*nx+sx] = shot_val[i];
   }  
}
void Source::BInjection(Vector *shot, Vector *u, llong sz, llong sx) {
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr();   
  float *__restrict__ u_val = ((dataInCoreFloat*)u->getDataContainer())->getFloatPtr();  

  for(llong i=0; i<nt;i++){
    u_val[i*nx*nz+sz*nx+sx] = shot_val[nt-i-1];
   }  
}

void Source::InjectionDipole(Vector *shot, Vector *u, llong sz, llong sx) {
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr();   
  float *__restrict__ u_val = ((dataInCoreFloat*)u->getDataContainer())->getFloatPtr();  

  for(llong i=0; i<nt;i++){
    u_val[i*nx*nz+sz*nx+sx] = shot_val[i];
    u_val[i*nx*nz+(sz+1)*nx+sx] = -1.*shot_val[i];
   }  
}

void Source::InjectionLine(Vector *shot, Vector *u, llong sz, llong sx) {
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr();   
  float *__restrict__ u_val = ((dataInCoreFloat*)u->getDataContainer())->getFloatPtr();  

  for(llong i=0; i<nt;i++){
    for(llong ix=0; ix<nx; ix++){
      u_val[i*nx*nz+sz*nx+ix] = shot_val[i];
    }
   }  
}


}

#endif