#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"

#include <omp.h>
#include "base/utilities.h"

#include "base/s_reg.h"
#include "base/hypercube.h"
#include "base/Vector.h"
#include "base/inCoreFloat.h"
#include "base/DataInCoreFloat.h"
#include "base/Operator.h"
#include "base/MultiOperator.h"

#include "Interpolation.h"
#include "CreateModel.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Prop.h"
#include "Imaging.h"
#include "Engine.h"
#include "NCG.h"
extern "C" {
#include "sep3d.h"
}
using namespace std;
using namespace SEP;

/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/
extern "C"  int sep_get_data_axis_par(const char *r,int *ax,int *ff,float *o,float *d,char *l);




int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "         Full Waveform inversion.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(10);   // Initialize the parallel environment
  
  // Define dimension of interest
  int nx, nz, nt;
  int init;
  int sx, sz,ns,oxs,dxs;
  int boundary;
  float dx, dz, dt;
  float velocity, density, modulus;
  float fmax=15;
  int nb, bsize;
  nt = 400;
  // Get parameters
  getch("nb","d",&nb);    cout << "	nb = " << nb << endl;
  getch("bsize","d",&bsize);    cout << "	bsize = " << bsize << endl;
  getch("nx","d",&nx);    cout << "	nx = " << nx << endl;
  getch("nz","d",&nz);    cout << "	nz = " << nz << endl;
  getch("nt","d",&nt);    cout << "	nt = " << nt << endl;
  getch("boundary","d",&boundary);    cout << "	boundary = " << boundary << endl;
  getch("dx","f",&dx);    cout << "	dx = " << dx <<"[m]"<< endl;
  dz = dx;
  getch("sz","d",&sz);    cout << "	sz = " << sz << endl;
  getch("oxs","d",&oxs);  cout << "	oxs = " << oxs << endl;
  getch("dxs","d",&dxs);  cout << "	dxs = " << dxs << endl;
  getch("dt","f",&dt);    cout << "	dt = " << dt << "[ms]"<< endl;
  getch("ns","d",&ns);    cout << "	ns = " << ns << endl;
  getch("density","f",&density); cout << "	density = " << density << "[g/cm^3]" <<endl;
  getch("velocity","f",&velocity); cout << "	velocity = " << velocity << "[km/s]" <<endl;
  getch("init","d",&init);    cout << "	init = " << init << endl;
  modulus=velocity*velocity*density;  cout << "	modulus = " << modulus << "[GPa]" <<endl; 

  // Set up space 
  inCoreFloat1D sspace(nt);               // space for shot
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wave field cube
  inCoreFloat2D dspace(nt,nx);            // space for data space

  // Set up vectors
  Vector *K      = new Vector(&mspace);    K->zeroElement();      // migration velocity
  Vector *rho    = new Vector(&mspace);    rho->zeroElement();     // Velocity vector
  Vector *grad   = new Vector(&mspace);
  Vector *grad0  = new Vector(&mspace);
  Vector *dK      = new Vector(&mspace);    dK->zeroElement();      // migration velocity
  Vector *drho    = new Vector(&mspace);    drho->zeroElement();     // Velocity vector
  Vector *K0      = new Vector(&mspace);    K0->zeroElement();      // migration velocity
  Vector *Kold    = new Vector(&mspace);    Kold->zeroElement();      // migration velocity
  Vector *rho0    = new Vector(&mspace);    rho0->zeroElement();     // Velocity vector
  Vector *HK      = new Vector(&mspace);    HK->zeroElement();      // migration velocity
  Vector *Hrho    = new Vector(&mspace);    Hrho->zeroElement();     // Velocity vector
  Vector *shot   = new Vector(&sspace);    shot->zeroElement();     // shot profile

  // Read in wavelet and construct model
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  sreed("wavelet",shot_val,sizeof(float)*nt); 

  // Set up source 
  std::vector<int> sxv;
  std::vector<int> szv;
  for(int is=0; is < ns; is++){
    sxv.push_back(oxs+is*dxs);
  //  sxv.push_back(150);
    szv.push_back(sz);
  }

  // Set up model
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr(); 
  float *__restrict__ K0_val = ((dataInCoreFloat*)K0->getDataContainer())->getFloatPtr();
  float *__restrict__ rho0_val = ((dataInCoreFloat*)rho0->getDataContainer())->getFloatPtr();
  float *__restrict__ dK_val = ((dataInCoreFloat*)dK->getDataContainer())->getFloatPtr();
  float *__restrict__ drho_val = ((dataInCoreFloat*)drho->getDataContainer())->getFloatPtr();  
  // Set up Engine
  Engine *Run = new Engine(&mspace,&dspace,&uspace,shot,sxv,szv);
  Run->SetPropPar(nb, bsize, boundary, dt, dx, dz); 
  float *dataTrue=new float[ns*nx*nt];
  float *dataDirect=new float[ns*nx*nt]; 
  float *dataRes=new float[ns*nx*nt]; 
  double *dataResD=new double[ns*nx*nt];
  if( init==0){
    Models *CModel =  new Models(nz, nx, modulus, density);
    CModel->Model9(K,rho);
    CModel->Model8(K0,rho0); 
    srite("modelK",K_val,sizeof(float)*nz*nx);
    srite("modelr",rho_val,sizeof(float)*nz*nx); 
    srite("modelK0",K0_val,sizeof(float)*nz*nx);
    srite("modelr0",rho0_val,sizeof(float)*nz*nx);
    //CModel->ModelC(K,rho);
    //Run->SetModel(K,rho);
    //Run->CreateData();
    //Run->GetData(dataDirect);
    srite("dataDirect",dataDirect,sizeof(float)*ns*nt*nx);
    //for(int i=0; i < nt*nx*ns; i++) dataTrue[i] = dataTrue[i]-dataDirect[i];
    srite("dataTrue",dataTrue,sizeof(float)*ns*nt*nx);
    return 0;    
  }
  else{
    sreed("modelK",K_val,sizeof(float)*nz*nx);
    sreed("modelr",rho_val,sizeof(float)*nz*nx); 
    sreed("modelK0",K0_val,sizeof(float)*nz*nx);
    sreed("modelr0",rho0_val,sizeof(float)*nz*nx);
    sreed("dataTrue",dataTrue,sizeof(float)*ns*nt*nx); 
    srite("dataDirect",dataDirect,sizeof(float)*ns*nt*nx);   
  }


  // Collect data residual and inject data residual
  cout << "start to run " <<endl;
  Run->SetModel(K,rho);
  Run->CreateData();
  Run->GetData(dataTrue);
  Run->SetModel(K0,rho0);
  Run->CreateData();
  Run->GetData(dataRes);
  for(int i=0; i < nt*nx*ns; i++) dataRes[i] = dataRes[i]-dataTrue[i];
  double f0 = ObjFuc(dataRes,nt*nx*ns);
  double f;
  cout << "Initial model obj = " << f0 << endl;   
  Run->PutData(dataRes);
  
  // Iterative steepest descent
  double r0, r1, r2, r3,rmin;
  int Nfunc=0,niter=10,success=0;
  int nsteps;
  double alpha=5., objmin=1000000,alphamin;
  float *objs=new float[niter];
  double *alphas=new double[niter];
  float *Klong = new float[nx*nz*niter];
  
  LineSearch * LS = new LineSearch(&mspace);
  FuncGrad *func = new FuncGrad(&mspace,&mspace);
  func->SetData(dataRes,dataTrue,dataResD);
  func->SetOp(Run,drho,rho0, nt*ns*nx);
  
  // Test

/*
  // first use steepest descent
  double gold, gnew;
  func->Evaluate(K0,&f0,grad0);
  dK->linearCombo(0.,grad0,-1.);
  std::cout << "start from m norm = " << K0->innerProd(*K0) << std::endl;
  cout << " start from obj = "<< f0 << endl;
  gold = grad0->innerProd(*dK);
  for( int iter=0; iter < 20; iter++){
    cout << " Iteration :" << iter << ", start from obj = "<< f0 << endl;
    LS->run(&alpha,K,&f,grad, &nsteps, K0, f0, grad0,dK,func);    
    cout << "Line search finished in:" << nsteps << " steps, alpha = "<< alpha << ",value = " << f << endl;
    if(alpha < 0.001){
      std::cout << "    what the fuck is going on!" << std::endl; 
      alpha = 10.;
    }
    f0 = f;
    grad0->linearCombo(0.,*grad,1.);
    dK->linearCombo(0.,*grad,-1.);
    K0->linearCombo(0.,*K,1.);
    LS->SetPar(alpha);
  }
*/
  
  NCG *ncg = new NCG(&mspace);
  ncg->run(K, K0, func);
  std::cout << "restart ncg" << std::endl;
  //NCG *ncg2 = new NCG(&mspace);
  //K0->linearCombo(0.,*K,1.);
  //ncg2->run(K, K0, func); 

  float *__restrict__ HK_val = ((dataInCoreFloat*)HK->getDataContainer())->getFloatPtr();
  float *__restrict__ Hrho_val = ((dataInCoreFloat*)Hrho->getDataContainer())->getFloatPtr();   
  cout << "norm dK = " << dK->innerProd(*dK) << endl; 
  srite("data",dataRes,sizeof(float)*ns*nx*nt);
  srite("newK",K_val,sizeof(float)*nz*nx);
  srite("newr",rho_val,sizeof(float)*nz*nx);
  srite("Klong",Klong,sizeof(float)*nz*nx*niter);
  srite("objs",objs,sizeof(float)*niter);
  srite("alphas",alphas,sizeof(float)*niter);
  // Clean up

   std::cout << "========== Done, Cleaning up ==========" <<std::endl; 

  return 0;
}