#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"
#include <omp.h>
#include "base/utilities.h"

#include "base/Vector.h"
#include "base/inCoreFloat.h"
#include "base/DataInCoreFloat.h"
#include "base/Operator.h"
#include "base/MultiOperator.h"

#include "Interpolation.h"
#include "CreateModel.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Prop.h"
#include "Imaging.h"

using namespace std;
using namespace SEP;

/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/



int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "         Full Waveform inversion.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(4);   // Initialize the parallel environment
  
  // Define dimension of interest
  long long nx, nz, nt;
  long long sx, sz,ns,ox,dxs;
  long long boundary;
  float dx, dz, dt;
  float velocity, density, modulus;
  float fmax=15;
  nt = 1500;
  nx = 401;
  nz = 251;
  ns = 26;
  sx = 200;
  sz = 65;
  boundary = 4;

  density=2.;          // [g/cm^3]
  velocity=2.4;        // [km/s]
  dx = 10.; dz = 10.;  // [m]
  dt = 1.5;            // [ms]
  modulus=velocity*velocity*density; // [GPa]

  // Set up space 
  inCoreFloat1D sspace(nt);               // space for shot
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wave field cube
  inCoreFloat2D dspace(nt,nx);         // space for data space
  inCoreFloat3D bigspace(ns,nz,nx);         // space for data space
  // Set up vectors
  Vector *K      = new Vector(&mspace);    K->zeroElement();      // migration velocity
  Vector *rho    = new Vector(&mspace);    rho->zeroElement();     // Velocity vector
  Vector *K0      = new Vector(&mspace);    K0->zeroElement();      // migration velocity
  Vector *rho0    = new Vector(&mspace);    rho0->zeroElement();     // Velocity vector
  Vector *dK      = new Vector(&mspace);    dK->zeroElement();      // migration velocity
  Vector *drho    = new Vector(&mspace);    drho->zeroElement();     // Velocity vector
  Vector *GradK  = new Vector(&mspace);    GradK->zeroElement();      // migration velocity
  Vector *Gradr  = new Vector(&mspace);    Gradr->zeroElement();     // Velocity vector
  Vector *BigGradK= new Vector(&bigspace);   BigGradK->zeroElement();
  Vector *vel    = new Vector(&mspace);    vel->zeroElement();      // migration velocity
  Vector *shot   = new Vector(&sspace);    shot->zeroElement();     // shot profile
  Vector *rec    = new Vector(&sspace);    rec->zeroElement();     // shot profile
  Vector *rec2   = new Vector(&sspace);    rec2->zeroElement();     // shot profile
  Vector *rec0   = new Vector(&sspace);    rec0->zeroElement();     // shot profile
  Vector *data   = new Vector(&dspace);    data->zeroElement();     // data
  Vector *us     = new Vector(&uspace);    us->zeroElement();       // Wave field
  Vector *u0     = new Vector(&uspace);    u0->zeroElement();       // Wave field
  Vector *u1     = new Vector(&uspace);    u1->zeroElement();       // Wave field
  Vector *v1     = new Vector(&uspace);    v1->zeroElement();       // Wave field  
  Vector *v2a     = new Vector(&uspace);    v2a->zeroElement();       // Wave field 
  Vector *v2b    = new Vector(&uspace);    v2b->zeroElement(); 
  Vector *utemp1 = new Vector(&uspace);    utemp1->zeroElement();       // Wave field
  Vector *GNHessK  = new Vector(&mspace);    GNHessK->zeroElement();      // migration velocity  
  Vector *HessK1  = new Vector(&mspace);    HessK1->zeroElement();      // migration velocity
  Vector *HessK2  = new Vector(&mspace);    HessK2->zeroElement();      // migration velocity
  float *__restrict__ rec_val = ((dataInCoreFloat*)rec->getDataContainer())->getFloatPtr(); 
  float *__restrict__ u0_val = ((dataInCoreFloat*)u0->getDataContainer())->getFloatPtr();   
  float *__restrict__ u1_val = ((dataInCoreFloat*)u1->getDataContainer())->getFloatPtr();  
  float *__restrict__ v1_val = ((dataInCoreFloat*)v1->getDataContainer())->getFloatPtr();
  float *__restrict__ v2a_val = ((dataInCoreFloat*)v2a->getDataContainer())->getFloatPtr();
  float *__restrict__ v2b_val = ((dataInCoreFloat*)v2b->getDataContainer())->getFloatPtr();
  float *__restrict__ utemp_val = ((dataInCoreFloat*)utemp1->getDataContainer())->getFloatPtr();
  float *__restrict__ us_val = ((dataInCoreFloat*)us->getDataContainer())->getFloatPtr();
  // Read in source wavelet and velocity field
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  sreed("wavelet",shot_val,sizeof(float)*nt); 


  // Set up model
  Models *CModel =  new Models(nz, nx, modulus, density);
  CModel->Model6(K,rho);
  CModel->Model2(K0,rho0);
  float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr();
  float *__restrict__ K0_val = ((dataInCoreFloat*)K0->getDataContainer())->getFloatPtr(); 
  float *__restrict__ dK_val = ((dataInCoreFloat*)dK->getDataContainer())->getFloatPtr(); 
  float *__restrict__ rho_val = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr(); 
  float *__restrict__ rho0_val = ((dataInCoreFloat*)rho0->getDataContainer())->getFloatPtr(); 
  float *__restrict__ BigK_val = ((dataInCoreFloat*)BigGradK->getDataContainer())->getFloatPtr();
  float *__restrict__ gk_val = ((dataInCoreFloat*)GradK->getDataContainer())->getFloatPtr(); 
  float *__restrict__ gr_val = ((dataInCoreFloat*)Gradr->getDataContainer())->getFloatPtr(); 
  float *__restrict__ vel_val = ((dataInCoreFloat*)vel->getDataContainer())->getFloatPtr(); 
  
  for(llong i=0; i<nx*nz; i++){
    vel_val[i] = sqrt(K_val[i]/rho_val[i]);
  }
  srite("vel",vel_val,sizeof(float)*nx*nz);
  for(llong i=0; i<nx*nz; i++){
    vel_val[i] = sqrt(K0_val[i]/rho0_val[i]);
  }
  srite("vel0",vel_val,sizeof(float)*nx*nz);
  srite("modelK",K_val,sizeof(float)*nx*nz); 
  srite("modelr",rho_val,sizeof(float)*nx*nz); 

  //return 0;

  // Check numerical condition
  cout << "    Dispersion: Vmin/(fmax*dx) = " << 1000.*velocity/fmax/dx <<", should be > 4"<<endl;
  cout << "    Courant:    Vmax*dt/dx     = " << velocity*dt/dx <<", should be < 0.717"<<endl;
 
  // Generate operators for wave progation
  Prop *Lmig = new Prop(&uspace,&uspace);    // propogation to generate data
  Prop *Lmig0= new Prop(&uspace,&uspace);    // propagation using migration velocity
  Lmig->SetPropPar(boundary,dt, dx, K);
  Lmig0->SetPropPar(boundary,dt, dx, K);
  Lmig->SetModel(K,rho);
  Lmig0->SetModel(K0,rho0);
  //float *__restrict__ K_val = ((dataInCoreFloat*)K->getDataContainer())->getFloatPtr(); 
  K->linearCombo(1.,*K0,-1.); dK->linearCombo(0.,*K,1.);
  rho->linearCombo(1.,*rho0,-1.); drho->linearCombo(0.,*rho,1.);
  // Inject source wavelet
  Source *Ss = new Source(nt, nz, nx);
  //Ss->Injection(shot, u1, sz, sx);
 
  SensorPad *SP = new SensorPad(&dspace, &uspace, sz);
  SP->DotTest(true);
  Gradient *Grad = new Gradient(u1, K, boundary, dt, dx);
  
  for(llong is=0; is < ns; is++){
    sx = 70+10*is;
    cout << "Working on sx = " <<sx <<" ,shot " << is << " out of " << ns << endl;

    us->zeroElement();
    Ss->Injection(shot, us, sz, sx);
    // Forward propagation for u field and data
    double t0 = timer();
    Lmig0->Forward(false,us,u0);           // u is computed 
    Lmig->Forward(false,us,utemp1);         // source -> ufield
    //  float *__restrict__ d_val = ((dataInCoreFloat*)utemp1->getDataContainer())->getFloatPtr();
    //  srite("wave",d_val,sizeof(float)*nz*nx*nt);
	utemp1->linearCombo(1.,*u0,-1.);
    cout << "    data norm = " << utemp1->innerProd(*utemp1) << endl;

    SP->SetShot(sx);
    SP->Adjoint(false,data,utemp1);  
    double t1 = timer(); 
    cout << "   Forward wave propagation, takes " << t1-t0 << "seconds. " <<endl;

    // Now Adjoint propagation for lambda
    t0 = timer();
    SP->Forward(false,data,utemp1);      
    Lmig0->Adjoint(false,v1,utemp1);   // v1 = lambda
  //  if(is==4) {
    //  float *__restrict__ d_val = ((dataInCoreFloat*)utemp1->getDataContainer())->getFloatPtr();
    //  srite("wave",d_val,sizeof(float)*nz*nx*nt);
  //  }
    t1 = timer(); 
    cout << "   Backward wave propagation, takes " << t1-t0 << "seconds. " <<endl;
    
    // Now compute gradient
    GradK->zeroElement();
    Grad->GradK(u0,v1,GradK);
    cout << "    Grad norm = " << GradK->innerProd(*GradK) << endl;
    llong shift = is*nx*nz;
    for(llong j=0; j < nx*nz; j++){
      BigK_val[j+shift] = gk_val[j];
      gr_val[j] += gk_val[j];
    }
  } 
 


  srite("GradKmore",BigK_val,sizeof(float)*nz*nx*ns);
  srite("GradK",gr_val,sizeof(float)*nz*nx);

  return 0;  
   std::cout << "========== Done, Cleaning up ==========" <<std::endl; 

  return 0;
}