#ifndef TAPER_H
#define TAPER_H 1
#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "base/Vector.h"
#include <cmath>
#include <omp.h>
#include <iostream>
typedef long long llong;
/*
Wave propagation module
Need to set velocity and taper before running
Assume source field ready to use.

 Notice: nx, nz are labeled in the opposite way.
 I am too lazy to change all the variables.
 It works. Use carefully
Created:       2015-01-15
Last modified: 2015-01-27

*/
extern "C" void TaperISPC(float *cur,float *next,float *tap, int bnx, int n1, int n2);

namespace SEP {

class Taper{
  
  public:
    Taper(Vector *vec, llong _bnz, llong _bnx, float _alpha){
      init(vec, _bnz, _bnx, _alpha);
    }
    ~Taper(){}
    void tape(Vector *model);
    void tapeISPC(Vector *cur, Vector *next){
     float *__restrict__ m1 = ((dataInCoreFloat*)cur->getDataContainer())->getFloatPtr();
      float *__restrict__ m2 = ((dataInCoreFloat*)next->getDataContainer())->getFloatPtr();    
     float *__restrict__ tap = ((dataInCoreFloat*)taper->getDataContainer())->getFloatPtr();    
      TaperISPC(m1,m2,tap,bnx,nx,nz);
    }
  private:
    void init(Vector *vec, llong _bnz, llong _bnx, float _alpha);
    llong nz,nx,nt, bnz,bnx;
    float alpha,pi;
    Vector *taper;   
    inCoreFloat *mspace;


};



void Taper::init(Vector *vec, llong _bnz, llong _bnx, float _alpha) {
  taper = new Vector(vec);
  mspace = (inCoreFloat *)vec->getSpace();
  std::vector<long long> in_dims = mspace->get_ns();
  nx = in_dims[0]; 
  nz = in_dims[1]; 
  bnz = _bnz; bnx = _bnx; alpha = _alpha;
  pi = 3.1415926535;
  float *__restrict__ tap = ((dataInCoreFloat*)taper->getDataContainer())->getFloatPtr(); 
  
  // left and right
  for(llong ix = 0; ix < bnx; ix++) {
    for(llong iz = bnz; iz < nz-bnz; iz++){
      double taps = 1.*(bnx-ix)/(1.0*bnx);
      taps = alpha + (1.-alpha)*cos(pi*taps);
      tap[ix*nz+iz] = taps;
      tap[(nx-ix-1)*nz+iz] = taps;
    }
  }
  // up and bottom
  for(llong iz = 0; iz < bnz; iz++) {
    for(llong ix = bnx; ix < nx-bnx; ix++){
      double taps=1.*(bnz-iz)/(1.0*bnz);
      taps = alpha + (1.-alpha)*cos(pi*taps);
      tap[ix*nz+iz] = taps;
      tap[(nx-ix-1)*nz+nz-iz-1] = taps;
    }
  }
  // Corner
  for(llong ix=0; ix < bnx; ix++){
    for(llong iz=0; iz < bnz; iz++){
      double temp = 1.00*(bnx-ix)*(bnx-ix)/(1.*bnx*bnx)+1.00*(bnz-iz)*(bnz-iz)/(1.*bnz*bnz);
      temp = sqrt(temp);
      float temp2 = alpha + (1.-alpha)*cos(pi*temp);
      tap[ix*nz+iz] = temp2;
      tap[ix*nz+nz-iz-1] = temp2;
      tap[(nx-ix-1)*nz+iz] = temp2;
      tap[(nx-ix-1)*nz+nz-iz-1] = temp2;
    }
  }
  
  for(llong ix=bnx; ix < nx-bnx; ix++){
    for(llong iz=bnz; iz<nz-bnz; iz++){
      tap[ix*nz+iz] = 1.;
    }
  }
 double diff=0.;

}

void Taper::tape(Vector *model) {

  float *__restrict__ tap = ((dataInCoreFloat*)taper->getDataContainer())->getFloatPtr();
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();

  llong nshift = (nx-bnx)*nz;
  llong ileft, iright;
  llong i, j, midshift, midshift2;
  //std::cout << "	tapering "<<std::endl;
  #pragma omp parallel for private(ileft,iright)
  for(ileft = 0; ileft < bnx*nz; ileft++){
    iright = nshift + ileft; 
    m[ileft] = m[ileft]*tap[ileft];
    m[iright] = m[iright]*tap[iright];
  }

  llong ix,iz;
  #pragma omp parallel for private(ix,iz,midshift,midshift2,i,j)
  for(ix=bnx; ix < nx-bnx; ix++) {
     midshift = ix*nz;
     midshift2= ix*nz+nz - bnz;
     for(iz=0; iz<bnz; iz++){
      i = midshift + iz;
      m[i] = m[i]*tap[i];
      j =  midshift2+ iz;
      m[j] = m[j]*tap[j];
    } 
  
  
  }

}




}

#endif