#ifndef GRAD_H
#define GRAD_H 1

#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "Clean.h"
#include <omp.h>
typedef long long llong;
#define C1 0.125
#define C2 0.25
/*
	Created by Yinbin Ma  
	4th order Laplacian in 2 d.
	Use carefully and have fun.
	
	First created: 2015-01-01
	Last Modified: 2015-02-28
*/	

namespace SEP {

class GradX: public Operator{
  public:
    GradX(Space *mod, Space *data, llong _nb){
      setDomain(mod);
      setRange(data);
      nb = _nb;
      init();
    }
    ~GradX(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    private:
    void init();
    float a0, a1, a2;
    llong n1, n2;
    llong i1, i2, position;
    llong nb;
    float temp0,temp1,temp2;
    CleanVec Zero;

};

void GradX::init() {
  a0 = 0.5;
  a1 = 9./8.;
  a2 = -1./24.;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n1 = in_dims[0];
  n2 = in_dims[1];
}

bool GradX::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, position;
 //   #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-2; i2++){
      position = i1*n2+i2;
      d[position] += -1.*a2*m[position-1];   
      d[position] += -1.*a1*m[position];   
      d[position] += a1*m[position+1];
      d[position] += a2*m[position+2];

    }
  }



  return 0;
}

bool GradX::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, position;
 //   #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-2; i2++){
      position = i1*n2+i2;
      m[position-1] += -1.*a2*d[position];   
      m[position] += -1.*a1*d[position];   
      m[position+1] += a1*d[position];
      m[position+2] += a2*d[position];

    }
  }

  return 0;
}
class GradZ: public Operator{
  public:
    GradZ(Space *mod, Space *data, llong _nb){
      setDomain(mod);
      setRange(data);
      nb = _nb;
      init();
    }
    ~GradZ(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    private:
    void init();
    float a0, a1, a2;
    llong n1, n2;
    llong i1, i2, position;
    llong nb;
    float temp0,temp1,temp2;
    CleanVec Zero;

};

void GradZ::init() {
  a0 = 0.5;
  a1 = 9./8.;
  a2 = -1./24.;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n1 = in_dims[0];
  n2 = in_dims[1];
}

bool GradZ::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
    // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, position;
   // #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-2; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      position = i1*n2+i2;
      d[position] += -1.*a2*m[position-nx];   
      d[position] += -1.*a1*m[position];   
      d[position] += a1*m[position+nx];
      d[position] += a2*m[position+2*nx];

    }
  }

  return 0;
}

bool GradZ::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, position;
  for(i1 = nb; i1 < n1-nb-2; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      position = i1*n2+i2;
      m[position-nx] += -1.*a2*d[position];   
      m[position] += -1.*a1*d[position];   
      m[position+nx] += a1*d[position];
      m[position+2*nx] += a2*d[position];

    }
  }

  return 0;
}

class GradX2: public Operator{
  public:
    GradX2(Space *mod, Space *data, llong _nb){
      setDomain(mod);
      setRange(data);
      nb = _nb;
      init();
    }
    ~GradX2(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    private:
    void init();
    float a0, a1, a2;
    llong n1, n2;
    llong i1, i2, position;
    llong nb;
    float temp0,temp1,temp2;
    CleanVec Zero;

};

void GradX2::init() {
  a0 = 0.5;
  a1 = 9./8.;
  a2 = -1./24.;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n1 = in_dims[0];
  n2 = in_dims[1];
}

bool GradX2::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, position;
  //  #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      position = i1*n2+i2;
      d[position] += -1.*a2*m[position-2];   
      d[position] += -1.*a1*m[position-1];   
      d[position] += a1*m[position];
      d[position] += a2*m[position+1];

    }
  }



  return 0;
}

bool GradX2::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, position;
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      position = i1*n2+i2;
      m[position-2] += -1.*a2*d[position];   
      m[position-1] += -1.*a1*d[position];   
      m[position] += a1*d[position];
      m[position+1] += a2*d[position];

    }
  }

  return 0;
}
class GradZ2: public Operator{
  public:
    GradZ2(Space *mod, Space *data, llong _nb){
      setDomain(mod);
      setRange(data);
      nb = _nb;
      init();
    }
    ~GradZ2(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    private:
    void init();
    float a0, a1, a2;
    llong n1, n2;
    llong i1, i2, position;
    llong nb;
    float temp0,temp1,temp2;
    CleanVec Zero;

};

void GradZ2::init() {
  a0 = 0.5;
  a1 = 9./8.;
  a2 = -1./24.;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n1 = in_dims[0];
  n2 = in_dims[1];
}

bool GradZ2::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
    // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, position;
  //  #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      position = i1*n2+i2;
      d[position] += -1.*a2*m[position-nx*2];   
      d[position] += -1.*a1*m[position-nx];   
      d[position] += a1*m[position];
      d[position] += a2*m[position+nx];

    }
  }

  return 0;
}

bool GradZ2::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,n1,n2);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  // n1=nz, n2=nx
  llong nz=n1, nx=n2;
  llong i1, i2, position;
   // #pragma omp parallel for
  for(i1 = nb; i1 < n1-nb-1; i1++){
    for(i2 = nb; i2 <n2-nb-1; i2++){
      position = i1*n2+i2;
      m[position-2*nx] += -1.*a2*d[position];   
      m[position-nx] += -1.*a1*d[position];   
      m[position] += a1*d[position];
      m[position+nx] += a2*d[position];

    }
  }

  return 0;
}


}
#endif