#ifndef TIMEMARCH_H
#define TIMEMARCH_H 1

#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "Laplacian.h"
#include "Interpolation.h"
#include "Clean.h"
#include "Grad.h"
#include "Taper.h"
typedef long long llong;


/*
	Created by Yinbin Ma  
	Time marching operator to compute Unext = F(Ucur,Uprev) without taper/source
	Use carefully and have fun.
	Warning: This code does not clean *next, you may want to clean it before calling 
	         FMarch/BMarch
	First created: 2015-01-01
	Last Modified: 2015-02-28
*/	


namespace SEP {

class TimeMarch{
  public:
    TimeMarch(llong _nb, float dt, float dx, Vector *K){
      nb = _nb;
      init(dt, dx, K);
    }
    void SetModel(Vector *_K, Vector *_rho){
      //std::cout <<"	Reset Modulus and density.  "<<std::endl;
      KTX->linearCombo(0.,*_K,dtdx2);
      float *__restrict__ temp_val = ((dataInCoreFloat*)temp->getDataContainer())->getFloatPtr();
      float *__restrict__ rho_val = ((dataInCoreFloat*)_rho->getDataContainer())->getFloatPtr();

	  for(long long position=0; position<nx*nz;position++){
		temp_val[position] = 1./rho_val[position];
	  }

	  interpX->Forward(false,temp,mrhox);   // perturbation on the staggered grid
	  interpZ->Forward(false,temp,mrhoz);   // perturbation on the staggered grid
   
/*
        for(long long position=0; position<nx*nz;position++){
          mrhoz_val[position] = 1./rho_val[position];
          mrhox_val[position] = 1./rho_val[position];
        }
*/

/*
      for(long long iz=0; iz<nz;iz++){
        for(long long ix=0; ix<nx-1;ix++){
          long long position = iz*nx+ix;
          mrhox_val[position] = (1./rho_val[position]+1./rho_val[position+1])/2;
        }  
      } 
      for(long long iz=0; iz<nz;iz++){
          long long position = iz*nx+nx-1;
          mrhox_val[position] = (1./rho_val[position]+1./rho_val[position])/2;
      }  
      for(long long iz=0; iz<nz-1;iz++){
        for(long long ix=0; ix<nx;ix++){
          long long position = iz*nx+ix;
          mrhoz_val[position] = (1./rho_val[position]+1./rho_val[position+nx])/2;
        }  
      } 
        for(long long ix=0; ix<nx;ix++){
          long long position = nz*(nx-1)+ix;
          mrhoz_val[position] = (1./rho_val[position]+1./rho_val[position])/2;
        }
*/
    }
    ~TimeMarch(){}
    bool FMarch(Vector *pre, Vector *cur, Vector *next);
    bool BMarch(Vector *pre, Vector *cur, Vector *next);
  private:
    void init(float _dt, float _dx, Vector *vel);
    inCoreFloat *space;
    Vector *KTX, *mrhox,*mrhoz;
    Vector *temp, *tempX, *tempZ;
    llong nz, nx, nb;
    float dx, dt, dx2,dtdx2;
    CleanVec Zero;
    GradX *DX;
    GradZ *DZ;
    GradX2 *DX2;
    GradZ2 *DZ2;
    Laplacian4 *L;
	IntX *interpX;
	IntZ *interpZ;

};

void TimeMarch::init(float _dt, float _dx, Vector *K) {
  dt = _dt; dx = _dx;
  dx2 = dx*dx;
  KTX = new Vector(K);
  mrhox = new Vector(K);
  mrhoz = new Vector(K);
  temp = new Vector(K);
  tempX = new Vector(K);
  tempZ = new Vector(K);
  
  space = (inCoreFloat *)(KTX->getSpace());
  DX = new GradX(space,space, nb);
  DZ = new GradZ(space,space, nb);
  DX2 = new GradX2(space,space, nb);
  DZ2 = new GradZ2(space,space, nb);
  interpX = new IntX(space,space);
  interpZ = new IntZ(space,space);

  L = new Laplacian4(space,space);
  std::vector<long long> in_dims = space->get_ns();
  nz = in_dims[0];
  nx = in_dims[1];
  float *__restrict__ alpha = ((dataInCoreFloat*)KTX->getDataContainer())->getFloatPtr();
  dtdx2 = 1.*(dt/dx)*(dt/dx);
  for(long long i=0; i<nx*nz;i++){
    alpha[i] = alpha[i]*dtdx2;
  }     // vtx = (dt/dx)^2 v^2
 //  std::cout << "nx = "<<nx <<", nz = " << nz <<std::endl;
}

bool TimeMarch::FMarch(Vector *pre, Vector *cur, Vector *next){  

  float *__restrict__ v1 = ((dataInCoreFloat*)pre->getDataContainer())->getFloatPtr();
  float *__restrict__ v2 = ((dataInCoreFloat*)cur->getDataContainer())->getFloatPtr(); 
  float *__restrict__ v3 = ((dataInCoreFloat*)next->getDataContainer())->getFloatPtr();
  float *__restrict__ t = ((dataInCoreFloat*)temp->getDataContainer())->getFloatPtr();
  float *__restrict__ tx = ((dataInCoreFloat*)tempX->getDataContainer())->getFloatPtr();
  float *__restrict__ tz = ((dataInCoreFloat*)tempZ->getDataContainer())->getFloatPtr();
  float *__restrict__ mrhox_val = ((dataInCoreFloat*)mrhox->getDataContainer())->getFloatPtr();
  float *__restrict__ mrhoz_val = ((dataInCoreFloat*)mrhoz->getDataContainer())->getFloatPtr();
  float *__restrict__ Ks = ((dataInCoreFloat*)KTX->getDataContainer())->getFloatPtr();


  DX->Forward(false,cur,temp);
  for(llong i=0; i <nx*nz; i++){
    t[i] = t[i]*mrhox_val[i];
  }
  DX2->Forward(false,temp,tempX);
  //DX->Adjoint(false,tempX,temp);
  
  DZ->Forward(false,cur,temp);
  #pragma omp parallel for
  for(llong i=0; i <nx*nz; i++){
    t[i] = t[i]*mrhoz_val[i];
  }
  DZ2->Forward(false,temp,tempZ);  
  //DZ->Adjoint(false,tempZ,temp);
  // std::cout <<"KTX = "<< Ks[100*nx+100] <<std::endl;
  // compute u_cur dependent term
  #pragma omp parallel for
  for(llong i=0; i <nx*nz; i++){
    v3[i] = v2[i]*2.+Ks[i]*(tx[i]+tz[i])-v1[i];
  }  

}

bool TimeMarch::BMarch(Vector *pre, Vector *cur, Vector *next){
  float *__restrict__ v1 = ((dataInCoreFloat*)pre->getDataContainer())->getFloatPtr();
  float *__restrict__ v2 = ((dataInCoreFloat*)cur->getDataContainer())->getFloatPtr(); 
  float *__restrict__ v3 = ((dataInCoreFloat*)next->getDataContainer())->getFloatPtr();
  float *__restrict__ t = ((dataInCoreFloat*)temp->getDataContainer())->getFloatPtr();
  float *__restrict__ tx = ((dataInCoreFloat*)tempX->getDataContainer())->getFloatPtr();
  float *__restrict__ tz = ((dataInCoreFloat*)tempZ->getDataContainer())->getFloatPtr();
  float *__restrict__ mrhox_val = ((dataInCoreFloat*)mrhox->getDataContainer())->getFloatPtr();
  float *__restrict__ mrhoz_val = ((dataInCoreFloat*)mrhoz->getDataContainer())->getFloatPtr();
  float *__restrict__ Ks = ((dataInCoreFloat*)KTX->getDataContainer())->getFloatPtr();

  //DX->Forward(false,cur,temp);
  DX2->Adjoint(false,temp,cur);
  #pragma omp parallel for
  for(llong i=0; i <nx*nz; i++){
    t[i] = t[i]*mrhox_val[i];
  }
  DX->Adjoint(false,tempX,temp);
    
  //DZ->Forward(false,cur,temp);
  DZ2->Adjoint(false,temp,cur);
  #pragma omp parallel for
  for(llong i=0; i <nx*nz; i++){
    t[i] = t[i]*mrhoz_val[i];
  }
  DZ->Adjoint(false,tempZ,temp);  

  // compute u_cur dependent term
  #pragma omp parallel for
  for(llong i=0; i <nx*nz; i++){
    v3[i] = v2[i]*2.+Ks[i]*(tx[i]+tz[i])-v1[i];
  }  

}
}
#endif